# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException
from airflow.providers.standard.version_compat import AIRFLOW_V_3_1_PLUS

if not AIRFLOW_V_3_1_PLUS:
    raise AirflowOptionalProviderFeatureException("Human in the loop functionality needs Airflow 3.1+.")

import asyncio
from collections.abc import AsyncIterator
from datetime import datetime
from typing import TYPE_CHECKING, Any, Literal, TypedDict
from uuid import UUID

from asgiref.sync import sync_to_async

from airflow.providers.common.compat.sdk import ParamValidationError
from airflow.sdk import Param
from airflow.sdk.definitions.param import ParamsDict
from airflow.sdk.execution_time.hitl import (
    HITLUser,
    get_hitl_detail_content_detail,
    update_hitl_detail_response,
)
from airflow.sdk.timezone import utcnow
from airflow.triggers.base import BaseTrigger, TriggerEvent


class HITLTriggerEventSuccessPayload(TypedDict, total=False):
    """Minimum required keys for a success Human-in-the-loop TriggerEvent."""

    chosen_options: list[str]
    params_input: dict[str, dict[str, Any]]
    responded_by_user: HITLUser | None
    responded_at: datetime
    timedout: bool


class HITLTriggerEventFailurePayload(TypedDict):
    """Minimum required keys for a failed Human-in-the-loop TriggerEvent."""

    error: str
    error_type: Literal["timeout", "unknown", "validation"]


class HITLTrigger(BaseTrigger):
    """A trigger that checks whether Human-in-the-loop responses are received."""

    def __init__(
        self,
        *,
        ti_id: UUID,
        options: list[str],
        params: dict[str, dict[str, Any]],
        defaults: list[str] | None = None,
        multiple: bool = False,
        timeout_datetime: datetime | None,
        poke_interval: float = 5.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ti_id = ti_id
        self.poke_interval = poke_interval

        self.options = options
        self.multiple = multiple
        self.defaults = defaults
        self.timeout_datetime = timeout_datetime

        self.params = ParamsDict(
            {
                k: Param(
                    v.pop("value"),
                    **v,
                )
                if HITLTrigger._is_param(v)
                else Param(v)
                for k, v in params.items()
            },
        )

    @staticmethod
    def _is_param(value: Any) -> bool:
        return isinstance(value, dict) and all(key in value for key in ("description", "schema", "value"))

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize HITLTrigger arguments and classpath."""
        return (
            "airflow.providers.standard.triggers.hitl.HITLTrigger",
            {
                "ti_id": self.ti_id,
                "options": self.options,
                "defaults": self.defaults,
                "params": {k: self.params.get_param(k).serialize() for k in self.params},
                "multiple": self.multiple,
                "timeout_datetime": self.timeout_datetime,
                "poke_interval": self.poke_interval,
            },
        )

    async def _handle_timeout(self) -> TriggerEvent:
        """Handle HITL timeout logic and yield appropriate event."""
        resp = await sync_to_async(get_hitl_detail_content_detail)(ti_id=self.ti_id)

        # Case 1: Response arrived just before timeout
        if resp.response_received and resp.chosen_options:
            if TYPE_CHECKING:
                assert resp.responded_by_user is not None
                assert resp.responded_at is not None

            chosen_options_list = list(resp.chosen_options or [])
            self.log.info(
                "[HITL] responded_by=%s (id=%s) options=%s at %s (timeout fallback skipped)",
                resp.responded_by_user.name,
                resp.responded_by_user.id,
                chosen_options_list,
                resp.responded_at,
            )
            return TriggerEvent(
                HITLTriggerEventSuccessPayload(
                    chosen_options=chosen_options_list,
                    params_input=resp.params_input or {},
                    responded_at=resp.responded_at,
                    responded_by_user=HITLUser(
                        id=resp.responded_by_user.id,
                        name=resp.responded_by_user.name,
                    ),
                    timedout=False,
                )
            )

        # Case 2: No defaults defined → failure
        if self.defaults is None:
            return TriggerEvent(
                HITLTriggerEventFailurePayload(
                    error="The timeout has passed, and the response has not yet been received.",
                    error_type="timeout",
                )
            )

        # Case 3: Timeout fallback to default
        resp = await sync_to_async(update_hitl_detail_response)(
            ti_id=self.ti_id,
            chosen_options=self.defaults,
            params_input=self.params.dump(),
        )
        if TYPE_CHECKING:
            assert resp.responded_at is not None

        self.log.info(
            "[HITL] timeout reached before receiving response, fallback to default %s",
            self.defaults,
        )
        return TriggerEvent(
            HITLTriggerEventSuccessPayload(
                chosen_options=self.defaults,
                params_input=self.params.dump(),
                responded_by_user=None,
                responded_at=resp.responded_at,
                timedout=True,
            )
        )

    async def _handle_response(self):
        """Check if HITL response is ready and yield success if so."""
        resp = await sync_to_async(get_hitl_detail_content_detail)(ti_id=self.ti_id)
        if TYPE_CHECKING:
            assert resp.responded_by_user is not None
            assert resp.responded_at is not None

        if not (resp.response_received and resp.chosen_options):
            return None

        # validate input
        if params_input := resp.params_input:
            try:
                for key, value in params_input.items():
                    self.params[key] = value
            except ParamValidationError as err:
                return TriggerEvent(
                    HITLTriggerEventFailurePayload(
                        error=str(err),
                        error_type="validation",
                    )
                )

        chosen_options_list = list(resp.chosen_options or [])
        self.log.info(
            "[HITL] responded_by=%s (id=%s) options=%s at %s",
            resp.responded_by_user.name,
            resp.responded_by_user.id,
            chosen_options_list,
            resp.responded_at,
        )
        return TriggerEvent(
            HITLTriggerEventSuccessPayload(
                chosen_options=chosen_options_list,
                params_input=params_input or {},
                responded_at=resp.responded_at,
                responded_by_user=HITLUser(
                    id=resp.responded_by_user.id,
                    name=resp.responded_by_user.name,
                ),
                timedout=False,
            )
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Loop until the Human-in-the-loop response received or timeout reached."""
        while True:
            if self.timeout_datetime and self.timeout_datetime < utcnow():
                event = await self._handle_timeout()
                yield event
                return

            event = await self._handle_response()
            if event:
                yield event
                return

            await asyncio.sleep(self.poke_interval)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow import DAG
from airflow.providers.common.compat.sdk import timezone
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.operators.trigger_dagrun import TriggerDagRunOperator
from airflow.providers.standard.sensors.external_task import ExternalTaskSensor

with DAG(
    dag_id="example_external_task",
    start_date=timezone.datetime(2022, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example", "async", "core"],
) as dag:
    start = EmptyOperator(task_id="start")

    # [START howto_external_task_async_sensor]
    external_task_sensor = ExternalTaskSensor(
        task_id="parent_task_sensor",
        external_task_id="child_task",
        external_dag_id="child_dag",
        deferrable=True,
    )
    # [END howto_external_task_async_sensor]

    trigger_child_task = TriggerDagRunOperator(
        task_id="trigger_child_task",
        trigger_dag_id="child_dag",
        allowed_states=[
            "success",
            "failed",
        ],
        logical_date="{{ logical_date }}",
        poke_interval=5,
        reset_dag_run=True,
        wait_for_completion=True,
    )

    end = EmptyOperator(task_id="end")

    start >> [trigger_child_task, external_task_sensor] >> end



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import typing
from collections.abc import Collection
from typing import Any

from asgiref.sync import sync_to_async
from sqlalchemy import func, select

from airflow.models import DagRun
from airflow.providers.standard.utils.sensor_helper import _get_count
from airflow.providers.standard.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.triggers.base import BaseTrigger, TriggerEvent

if typing.TYPE_CHECKING:
    from datetime import datetime

    from sqlalchemy.orm import Session

    from airflow.utils.state import DagRunState


class WorkflowTrigger(BaseTrigger):
    """
    A trigger to monitor tasks, task group and dag execution in Apache Airflow.

    :param external_dag_id: The ID of the external dag.
    :param run_ids: A list of run ids for the external dag.
    :param external_task_ids: A collection of external task IDs to wait for.
    :param external_task_group_id: The ID of the external task group to wait for.
    :param failed_states: States considered as failed for external tasks.
    :param skipped_states: States considered as skipped for external tasks.
    :param allowed_states: States considered as successful for external tasks.
    :param poke_interval: The interval (in seconds) for poking the external tasks.
    :param soft_fail: If True, the trigger will not fail the entire dag on external task failure.
    :param logical_dates: A list of logical dates for the external dag.
    """

    def __init__(
        self,
        external_dag_id: str,
        run_ids: list[str] | None = None,
        execution_dates: list[datetime] | None = None,
        logical_dates: list[datetime] | None = None,
        external_task_ids: typing.Collection[str] | None = None,
        external_task_group_id: str | None = None,
        failed_states: Collection[str] | None = None,
        skipped_states: Collection[str] | None = None,
        allowed_states: Collection[str] | None = None,
        poke_interval: float = 2.0,
        soft_fail: bool = False,
        **kwargs,
    ):
        self.external_dag_id = external_dag_id
        self.external_task_ids = external_task_ids
        self.external_task_group_id = external_task_group_id
        self.failed_states = failed_states
        self.skipped_states = skipped_states
        self.allowed_states = allowed_states
        self.run_ids = run_ids
        self.poke_interval = poke_interval
        self.soft_fail = soft_fail
        self.execution_dates = execution_dates
        self.logical_dates = logical_dates
        super().__init__(**kwargs)

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize the trigger param and module path."""
        data: dict[str, typing.Any] = {
            "external_dag_id": self.external_dag_id,
            "external_task_ids": self.external_task_ids,
            "external_task_group_id": self.external_task_group_id,
            "failed_states": self.failed_states,
            "skipped_states": self.skipped_states,
            "allowed_states": self.allowed_states,
            "poke_interval": self.poke_interval,
            "soft_fail": self.soft_fail,
        }
        if AIRFLOW_V_3_0_PLUS:
            data["run_ids"] = self.run_ids
            data["logical_dates"] = self.logical_dates
        else:
            data["execution_dates"] = self.execution_dates

        return "airflow.providers.standard.triggers.external_task.WorkflowTrigger", data

    async def run(self) -> typing.AsyncIterator[TriggerEvent]:
        """Check periodically tasks, task group or dag status."""
        if AIRFLOW_V_3_0_PLUS:
            get_count_func = self._get_count_af_3
            run_id_or_dates = (self.run_ids or self.logical_dates) or []
        else:
            get_count_func = self._get_count
            run_id_or_dates = self.execution_dates or []

        while True:
            if self.failed_states:
                failed_count = await get_count_func(self.failed_states)
                if failed_count > 0:
                    yield TriggerEvent({"status": "failed"})
                    return

            if self.skipped_states:
                skipped_count = await get_count_func(self.skipped_states)
                if skipped_count > 0:
                    yield TriggerEvent({"status": "skipped"})
                    return
            allowed_count = await get_count_func(self.allowed_states)

            if allowed_count == len(run_id_or_dates):
                yield TriggerEvent({"status": "success"})
                return
            self.log.info("Sleeping for %s seconds", self.poke_interval)
            await asyncio.sleep(self.poke_interval)

    async def _get_count_af_3(self, states: Collection[str] | None) -> int:
        from airflow.providers.standard.utils.sensor_helper import _get_count_by_matched_states
        from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance

        if self.external_task_ids:
            count = await sync_to_async(RuntimeTaskInstance.get_ti_count)(
                dag_id=self.external_dag_id,
                task_ids=list(self.external_task_ids),
                logical_dates=self.logical_dates,
                run_ids=self.run_ids,
                states=list(states) if states else None,
            )
            return int(count / len(self.external_task_ids))
        if self.external_task_group_id:
            run_id_task_state_map = await sync_to_async(RuntimeTaskInstance.get_task_states)(
                dag_id=self.external_dag_id,
                task_group_id=self.external_task_group_id,
                logical_dates=self.logical_dates,
                run_ids=self.run_ids,
            )
            count = await sync_to_async(_get_count_by_matched_states)(
                run_id_task_state_map=run_id_task_state_map,
                states=states or [],
            )
            return count
        count = await sync_to_async(RuntimeTaskInstance.get_dr_count)(
            dag_id=self.external_dag_id,
            logical_dates=self.logical_dates,
            run_ids=self.run_ids,
            states=list(states) if states else None,
        )
        return count

    @sync_to_async
    def _get_count(self, states: Collection[str] | None) -> int:
        """
        Get the count of records against dttm filter and states. Async wrapper for _get_count.

        :param states: task or dag states
        :return The count of records.
        """
        return _get_count(
            dttm_filter=self.run_ids if AIRFLOW_V_3_0_PLUS else self.execution_dates,
            external_task_ids=self.external_task_ids,
            external_task_group_id=self.external_task_group_id,
            external_dag_id=self.external_dag_id,
            states=states,
        )


class DagStateTrigger(BaseTrigger):
    """
    Waits asynchronously for a dag to complete for a specific run_id.

    :param dag_id: The dag_id that contains the task you want to wait for
    :param states: allowed states, default is ``['success']``
    :param run_ids: The run_id of dag run.
    :param poll_interval: The time interval in seconds to check the state.
        The default value is 5.0 sec.
    """

    def __init__(
        self,
        dag_id: str,
        states: list[DagRunState],
        run_ids: list[str] | None = None,
        execution_dates: list[datetime] | None = None,
        poll_interval: float = 5.0,
    ):
        super().__init__()
        self.dag_id = dag_id
        self.states = states
        self.run_ids = run_ids
        self.execution_dates = execution_dates
        self.poll_interval = poll_interval

    def serialize(self) -> tuple[str, dict[str, typing.Any]]:
        """Serialize DagStateTrigger arguments and classpath."""
        data = {
            "dag_id": self.dag_id,
            "states": self.states,
            "poll_interval": self.poll_interval,
            "run_ids": self.run_ids,
            "execution_dates": self.execution_dates,
        }

        return "airflow.providers.standard.triggers.external_task.DagStateTrigger", data

    async def run(self) -> typing.AsyncIterator[TriggerEvent]:
        """Check periodically if the dag run exists, and has hit one of the states yet, or not."""
        runs_ids_or_dates = 0
        if self.run_ids:
            runs_ids_or_dates = len(self.run_ids)
        elif self.execution_dates:
            runs_ids_or_dates = len(self.execution_dates)

        cls_path, data = self.serialize()

        if AIRFLOW_V_3_0_PLUS:
            data.update(  # update with {run_id: run_state} dict
                await self.validate_count_dags_af_3(runs_ids_or_dates_len=runs_ids_or_dates)
            )
            yield TriggerEvent((cls_path, data))
            return
        else:
            while True:
                num_dags = await self.count_dags()
                if num_dags == runs_ids_or_dates:
                    yield TriggerEvent((cls_path, data))
                    return
                await asyncio.sleep(self.poll_interval)

    async def validate_count_dags_af_3(self, runs_ids_or_dates_len: int = 0) -> dict[str, str]:
        from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance

        run_states: dict[str, str] = {}  # {run_id: run_state}
        while True:
            num_dags = await sync_to_async(RuntimeTaskInstance.get_dr_count)(
                dag_id=self.dag_id,
                run_ids=self.run_ids,
                states=self.states,  # type: ignore[arg-type]
                logical_dates=self.execution_dates,
            )
            if num_dags == runs_ids_or_dates_len:
                if isinstance(self.run_ids, list):
                    for run_id in self.run_ids:
                        state = await sync_to_async(RuntimeTaskInstance.get_dagrun_state)(
                            dag_id=self.dag_id,
                            run_id=run_id,
                        )
                        run_states[run_id] = state
                        return run_states
            await asyncio.sleep(self.poll_interval)

    if not AIRFLOW_V_3_0_PLUS:
        from airflow.utils.session import NEW_SESSION, provide_session  # type: ignore[misc]

        @sync_to_async
        @provide_session
        def count_dags(self, *, session: Session = NEW_SESSION) -> int:
            """Count how many dag runs in the database match our criteria."""
            _dag_run_date_condition = (
                DagRun.run_id.in_(self.run_ids or [])
                if AIRFLOW_V_3_0_PLUS
                else DagRun.execution_date.in_(self.execution_dates)
            )
            stmt = (
                select(func.count())
                .select_from(DagRun)
                .where(
                    DagRun.dag_id == self.dag_id,
                    DagRun.state.in_(self.states),
                    _dag_run_date_condition,
                )
            )
            result = session.execute(stmt).scalar()
            return result or 0



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from unittest.mock import MagicMock, Mock

import pytest
from sqlalchemy import select

from airflow.models.taskinstance import TaskInstance as TI
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.utils import timezone
from airflow.utils.state import State
from airflow.utils.types import DagRunType

from tests_common.test_utils.db import clear_db_dags, clear_db_runs
from tests_common.test_utils.taskinstance import create_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

pytestmark = pytest.mark.db_test

if AIRFLOW_V_3_0_PLUS:
    from airflow.models.dag_version import DagVersion
    from airflow.providers.common.compat.sdk import DownstreamTasksSkipped
    from airflow.providers.standard.utils.skipmixin import SkipMixin
    from airflow.sdk import task, task_group
    from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance
else:
    from airflow.decorators import task, task_group  # type: ignore[attr-defined,no-redef]
    from airflow.models.skipmixin import SkipMixin

DEFAULT_DATE = timezone.datetime(2016, 1, 1)
DEFAULT_DAG_RUN_ID = "test1"


class TestSkipMixin:
    @staticmethod
    def clean_db():
        clear_db_dags()
        clear_db_runs()

    def setup_method(self):
        self.clean_db()

    def teardown_method(self):
        self.clean_db()

    def test_skip(self, dag_maker, session, time_machine):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        time_machine.move_to(now, tick=False)
        with dag_maker("dag"):
            tasks = [EmptyOperator(task_id="task")]

        if AIRFLOW_V_3_0_PLUS:
            dag_run = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                logical_date=now,
                state=State.FAILED,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                SkipMixin().skip(ti=dag_run.get_task_instance("task"), tasks=tasks)

            assert exc_info.value.tasks == ["task"]
        else:
            from airflow import settings

            dag_run = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                execution_date=now,
                state=State.FAILED,
            )
            session = settings.Session()
            SkipMixin().skip(dag_run=dag_run, execution_date=now, tasks=tasks)

            session.scalar(
                select(TI).where(
                    TI.dag_id == "dag",
                    TI.task_id == "task",
                    TI.state == State.SKIPPED,
                    TI.start_date == now,
                    TI.end_date == now,
                )
            )

    def test_skip_none_tasks(self):
        if AIRFLOW_V_3_0_PLUS:
            assert SkipMixin().skip(ti=Mock(), tasks=[]) is None
        else:
            session = Mock()
            assert SkipMixin().skip(dag_run=None, execution_date=None, tasks=[]) is None
            assert not session.query.called
            assert not session.commit.called

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Airflow 2 had a different implementation")
    def test_skip__only_mapped_operators_passed(self):
        from airflow.sdk.definitions.mappedoperator import MappedOperator

        ti = Mock(map_index=2)
        assert SkipMixin().skip(ti=ti, tasks=[MagicMock(spec=MappedOperator)]) is None

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Airflow 2 had a different implementation")
    def test_skip__only_none_mapped_operators_passed(self):
        from airflow.sdk.definitions.mappedoperator import MappedOperator

        ti = Mock(map_index=-1)
        with pytest.raises(DownstreamTasksSkipped) as exc_info:
            SkipMixin().skip(ti=ti, tasks=[MagicMock(spec=MappedOperator, task_id="task")])
        assert exc_info.value.tasks == ["task"]

    @pytest.mark.parametrize(
        ("branch_task_ids", "expected_states"),
        [
            (None, {"task2": State.SKIPPED, "task3": State.SKIPPED}),
            ([], {"task2": State.SKIPPED, "task3": State.SKIPPED}),
        ],
        ids=["None", "empty-list"],
    )
    def test_skip_all_except__branch_task_ids_none(
        self, dag_maker, branch_task_ids, expected_states, session
    ):
        with dag_maker(
            "dag_test_skip_all_except",
            serialized=True,
        ):
            task1 = EmptyOperator(task_id="task1")
            task2 = EmptyOperator(task_id="task2")
            task3 = EmptyOperator(task_id="task3")

            task1 >> [task2, task3]
        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        if AIRFLOW_V_3_0_PLUS:
            ti1 = RuntimeTaskInstance.model_construct(
                dag_id=task1.dag_id,
                task_id=task1.task_id,
                task=task1,
                run_id=DEFAULT_DAG_RUN_ID,
            )
            ti1.__dict__["xcom_push"] = MagicMock()
        else:
            ti1 = TI(task1, run_id=DEFAULT_DAG_RUN_ID)

        if AIRFLOW_V_3_0_PLUS:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                SkipMixin().skip_all_except(ti=ti1, branch_task_ids=branch_task_ids)

            assert set(exc_info.value.tasks) == {("task2", -1), ("task3", -1)}
        else:
            ti2 = TI(task2, run_id=DEFAULT_DAG_RUN_ID)
            ti3 = TI(task3, run_id=DEFAULT_DAG_RUN_ID)

            SkipMixin().skip_all_except(ti=ti1, branch_task_ids=branch_task_ids)

            session.expire_all()

            def get_state(ti):
                ti.refresh_from_db()
                return ti.state

            executed_states = {"task2": get_state(ti2), "task3": get_state(ti3)}

            assert executed_states == expected_states

    @pytest.mark.parametrize(
        ("branch_task_ids", "expected_states"),
        [
            (["task2"], {"task2": State.NONE, "task3": State.SKIPPED}),
            (("task2",), {"task2": State.NONE, "task3": State.SKIPPED}),
            ("task2", {"task2": State.NONE, "task3": State.SKIPPED}),
        ],
        ids=["list-of-task-ids", "tuple-of-task-ids", "str-task-id"],
    )
    def test_skip_all_except__skip_task3(self, dag_maker, branch_task_ids, expected_states, session):
        with dag_maker(
            "dag_test_skip_all_except",
            serialized=True,
        ):
            task1 = EmptyOperator(task_id="task1")
            task2 = EmptyOperator(task_id="task2")
            task3 = EmptyOperator(task_id="task3")

            task1 >> [task2, task3]
        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        if AIRFLOW_V_3_0_PLUS:
            ti1 = RuntimeTaskInstance.model_construct(
                dag_id=task1.dag_id,
                task_id=task1.task_id,
                task=task1,
                run_id=DEFAULT_DAG_RUN_ID,
            )
            ti1.__dict__["xcom_push"] = MagicMock()
        else:
            ti1 = TI(task1, run_id=DEFAULT_DAG_RUN_ID)

        if AIRFLOW_V_3_0_PLUS:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                SkipMixin().skip_all_except(ti=ti1, branch_task_ids=branch_task_ids)

            assert set(exc_info.value.tasks) == {("task3", -1)}
        else:
            ti2 = TI(task2, run_id=DEFAULT_DAG_RUN_ID)
            ti3 = TI(task3, run_id=DEFAULT_DAG_RUN_ID)

            SkipMixin().skip_all_except(ti=ti1, branch_task_ids=branch_task_ids)

            session.expire_all()

            def get_state(ti):
                ti.refresh_from_db()
                return ti.state

            executed_states = {"task2": get_state(ti2), "task3": get_state(ti3)}

            assert executed_states == expected_states

    @pytest.mark.skipif(
        AIRFLOW_V_3_0_PLUS, reason="In Airflow 3, `NotPreviouslySkippedDep` is used for this case"
    )
    def test_mapped_tasks_skip_all_except(self, dag_maker):
        with dag_maker("dag_test_skip_all_except") as dag:

            @task
            def branch_op(k): ...

            @task_group
            def task_group_op(k):
                branch_a = EmptyOperator(task_id="branch_a")
                branch_b = EmptyOperator(task_id="branch_b")
                branch_op(k) >> [branch_a, branch_b]

            task_group_op.expand(k=[0, 1])

        def _create_task_instance(task_id, map_index):
            task = dag.get_task(task_id)
            if not AIRFLOW_V_3_0_PLUS:
                return TI(task, run_id=DEFAULT_DAG_RUN_ID, map_index=map_index)
            ti = RuntimeTaskInstance.model_construct(
                dag_id=dag.dag_id,
                task_id=task_id,
                task=task,
                run_id=DEFAULT_DAG_RUN_ID,
                map_index=map_index,
            )
            ti.__dict__["xcom_push"] = MagicMock()
            return ti

        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        branch_op_ti_0 = _create_task_instance("task_group_op.branch_op", map_index=0)
        branch_op_ti_1 = _create_task_instance("task_group_op.branch_op", map_index=1)
        branch_a_ti_0 = _create_task_instance("task_group_op.branch_a", map_index=0)
        branch_a_ti_1 = _create_task_instance("task_group_op.branch_a", map_index=1)
        branch_b_ti_0 = _create_task_instance("task_group_op.branch_b", map_index=0)
        branch_b_ti_1 = _create_task_instance("task_group_op.branch_b", map_index=1)

        SkipMixin().skip_all_except(ti=branch_op_ti_0, branch_task_ids="task_group_op.branch_a")
        SkipMixin().skip_all_except(ti=branch_op_ti_1, branch_task_ids="task_group_op.branch_b")

        def get_state(ti):
            ti.refresh_from_db()
            return ti.state

        assert get_state(branch_a_ti_0) == State.NONE
        assert get_state(branch_b_ti_0) == State.SKIPPED
        assert get_state(branch_a_ti_1) == State.SKIPPED
        assert get_state(branch_b_ti_1) == State.NONE

    def test_raise_exception_on_not_accepted_branch_task_ids_type(self, dag_maker):
        with dag_maker("dag_test_skip_all_except_wrong_type"):
            task = EmptyOperator(task_id="task")
        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        if AIRFLOW_V_3_0_PLUS:
            dag_version = DagVersion.get_latest_version(task.dag_id)
            ti1 = create_task_instance(task, run_id=DEFAULT_DAG_RUN_ID, dag_version_id=dag_version.id)
        else:
            ti1 = TI(task, run_id=DEFAULT_DAG_RUN_ID)
        error_message = (
            r"'branch_task_ids' must be either None, a task ID, or an Iterable of IDs, but got 'int'\."
        )
        with pytest.raises(AirflowException, match=error_message):
            SkipMixin().skip_all_except(ti=ti1, branch_task_ids=42)

    def test_raise_exception_on_not_accepted_iterable_branch_task_ids_type(self, dag_maker):
        with dag_maker("dag_test_skip_all_except_wrong_type"):
            task = EmptyOperator(task_id="task")
        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        if AIRFLOW_V_3_0_PLUS:
            dag_version = DagVersion.get_latest_version(task.dag_id)
            ti1 = create_task_instance(task, run_id=DEFAULT_DAG_RUN_ID, dag_version_id=dag_version.id)
        else:
            ti1 = TI(task, run_id=DEFAULT_DAG_RUN_ID)

        if AIRFLOW_V_3_0_PLUS:
            # Improved error message for Airflow 3.0+
            error_message = (
                r"Unable to branch to the specified tasks\. "
                r"The branching function returned invalid 'branch_task_ids': \{\(42, 'int'\)\}\. "
                r"Please check that your function returns an Iterable of valid task IDs that exist in your DAG\."
            )
        else:
            # Old error message for Airflow 2.x
            error_message = (
                r"'branch_task_ids' expected all task IDs are strings\. "
                r"Invalid tasks found: \{\(42, 'int'\)\}\."
            )

        with pytest.raises(AirflowException, match=error_message):
            SkipMixin().skip_all_except(ti=ti1, branch_task_ids=["task", 42])

    @pytest.mark.parametrize(
        "branch_task_ids",
        [
            pytest.param("task4", id="invalid-single-task"),
            pytest.param(["task2", "task4"], id="invalid-any-task-in-list"),
            pytest.param(["task5", "task4"], id="invalid-all-task-in-list"),
        ],
    )
    def test_raise_exception_on_not_valid_branch_task_ids(self, dag_maker, branch_task_ids):
        with dag_maker("dag_test_skip_all_except_wrong_type", serialized=True):
            task1 = EmptyOperator(task_id="task1")
            task2 = EmptyOperator(task_id="task2")
            task3 = EmptyOperator(task_id="task3")

            task1 >> [task2, task3]
        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)
        if AIRFLOW_V_3_0_PLUS:
            dag_version = DagVersion.get_latest_version(task1.dag_id)
            ti1 = create_task_instance(task1, run_id=DEFAULT_DAG_RUN_ID, dag_version_id=dag_version.id)
        else:
            ti1 = TI(task1, run_id=DEFAULT_DAG_RUN_ID)

        error_message = r"'branch_task_ids' must contain only valid task_ids. Invalid tasks found: .*"
        with pytest.raises(AirflowException, match=error_message):
            SkipMixin().skip_all_except(ti=ti1, branch_task_ids=branch_task_ids)

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Issue only exists in Airflow 3.x")
    def test_ensure_tasks_includes_sensors_airflow_3x(self, dag_maker):
        """Test that sensors (inheriting from airflow.sdk.BaseOperator) are properly handled by _ensure_tasks."""
        from airflow.providers.standard.utils.skipmixin import _ensure_tasks
        from airflow.sdk import BaseOperator as SDKBaseOperator
        from airflow.sdk.bases.sensor import BaseSensorOperator

        class DummySensor(BaseSensorOperator):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.timeout = 0
                self.poke_interval = 0

            def poke(self, context):
                return True

        with dag_maker("dag_test_sensor_skipping") as dag:
            regular_task = EmptyOperator(task_id="regular_task")
            sensor_task = DummySensor(task_id="sensor_task")
            downstream_task = EmptyOperator(task_id="downstream_task")

            regular_task >> [sensor_task, downstream_task]

        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)

        downstream_nodes = dag.get_task("regular_task").downstream_list
        task_list = _ensure_tasks(downstream_nodes)

        # Verify both the regular operator and sensor are included
        task_ids = [t.task_id for t in task_list]
        assert "sensor_task" in task_ids, "Sensor should be included in task list"
        assert "downstream_task" in task_ids, "Regular task should be included in task list"
        assert len(task_list) == 2, "Both tasks should be included"

        # Also verify that the sensor is actually an instance of the correct BaseOperator
        sensor_in_list = next((t for t in task_list if t.task_id == "sensor_task"), None)
        assert sensor_in_list is not None, "Sensor task should be found in list"
        assert isinstance(sensor_in_list, SDKBaseOperator), "Sensor should be instance of SDK BaseOperator"

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Integration test for Airflow 3.x sensor skipping")
    def test_skip_sensor_in_branching_scenario(self, dag_maker):
        """Integration test: verify sensors are properly skipped by branching operators in Airflow 3.x."""
        from airflow.sdk.bases.sensor import BaseSensorOperator

        # Create a dummy sensor for testing
        class DummySensor(BaseSensorOperator):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.timeout = 0
                self.poke_interval = 0

            def poke(self, context):
                return True

        with dag_maker("dag_test_branch_sensor_skipping") as dag:
            branch_task = EmptyOperator(task_id="branch_task")
            regular_task = EmptyOperator(task_id="regular_task")
            sensor_task = DummySensor(task_id="sensor_task")
            branch_task >> [regular_task, sensor_task]

        dag_maker.create_dagrun(run_id=DEFAULT_DAG_RUN_ID)

        ti_branch = RuntimeTaskInstance.model_construct(
            dag_id=dag.dag_id,
            task_id="branch_task",
            task=branch_task,
            run_id=DEFAULT_DAG_RUN_ID,
        )
        ti_branch.__dict__["xcom_push"] = MagicMock()

        # Test skipping the sensor (follow regular_task branch)
        with pytest.raises(DownstreamTasksSkipped) as exc_info:
            SkipMixin().skip_all_except(ti=ti_branch, branch_task_ids="regular_task")

        # Verify that the sensor task is properly marked for skipping
        skipped_tasks = set(exc_info.value.tasks)
        assert ("sensor_task", -1) in skipped_tasks, "Sensor task should be marked for skipping"

        # Test skipping the regular task (follow sensor_task branch)
        with pytest.raises(DownstreamTasksSkipped) as exc_info:
            SkipMixin().skip_all_except(ti=ti_branch, branch_task_ids="sensor_task")

        # Verify that the regular task is properly marked for skipping
        skipped_tasks = set(exc_info.value.tasks)
        assert ("regular_task", -1) in skipped_tasks, "Regular task should be marked for skipping"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import shutil
import tempfile
from datetime import timedelta

import pytest

try:
    from airflow.sdk.timezone import datetime
except ImportError:
    from airflow.utils.timezone import datetime  # type: ignore[no-redef]

from airflow.models.dag import DAG
from airflow.providers.common.compat.sdk import AirflowSensorTimeout, TaskDeferred
from airflow.providers.standard.sensors.filesystem import FileSensor
from airflow.providers.standard.triggers.file import FileTrigger

pytestmark = pytest.mark.db_test


TEST_DAG_ID = "unit_tests_file_sensor"
DEFAULT_DATE = datetime(2015, 1, 1)


class TestFileSensor:
    def setup_method(self):
        from airflow.providers.standard.hooks.filesystem import FSHook

        hook = FSHook()
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        dag = DAG(TEST_DAG_ID + "test_schedule_dag_once", schedule=timedelta(days=1), default_args=args)
        self.hook = hook
        self.dag = dag

    def test_simple(self):
        with tempfile.NamedTemporaryFile() as tmp:
            task = FileSensor(
                task_id="test",
                filepath=tmp.name[1:],
                fs_conn_id="fs_default",
                dag=self.dag,
                timeout=0,
            )
            task._hook = self.hook
            task.execute({})

    def test_file_in_nonexistent_dir(self):
        temp_dir = tempfile.mkdtemp()
        task = FileSensor(
            task_id="test",
            filepath=temp_dir[1:] + "/file",
            fs_conn_id="fs_default",
            dag=self.dag,
            timeout=0,
            poke_interval=1,
        )
        task._hook = self.hook
        try:
            with pytest.raises(AirflowSensorTimeout):
                task.execute({})
        finally:
            shutil.rmtree(temp_dir)

    def test_empty_dir(self):
        temp_dir = tempfile.mkdtemp()
        task = FileSensor(
            task_id="test",
            filepath=temp_dir[1:],
            fs_conn_id="fs_default",
            dag=self.dag,
            timeout=0,
            poke_interval=1,
        )
        task._hook = self.hook
        try:
            with pytest.raises(AirflowSensorTimeout):
                task.execute({})
        finally:
            shutil.rmtree(temp_dir)

    def test_file_in_dir(self):
        temp_dir = tempfile.mkdtemp()
        task = FileSensor(
            task_id="test",
            filepath=temp_dir[1:],
            fs_conn_id="fs_default",
            dag=self.dag,
            timeout=0,
        )
        task._hook = self.hook
        try:
            # `touch` the dir
            open(temp_dir + "/file", "a").close()
            task.execute({})
        finally:
            shutil.rmtree(temp_dir)

    def test_default_fs_conn_id(self):
        with tempfile.NamedTemporaryFile() as tmp:
            task = FileSensor(
                task_id="test",
                filepath=tmp.name[1:],
                dag=self.dag,
                timeout=0,
            )
            task._hook = self.hook
            task.execute({})

    def test_wildcard_file(self):
        suffix = ".txt"
        with tempfile.NamedTemporaryFile(suffix=suffix) as tmp:
            fileglob = os.path.join(os.path.dirname(tmp.name), "*" + suffix)
            task = FileSensor(
                task_id="test",
                filepath=fileglob,
                fs_conn_id="fs_default",
                dag=self.dag,
                timeout=0,
            )
            task._hook = self.hook
            task.execute({})

    def test_wildcard_empty_directory(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with tempfile.TemporaryDirectory(suffix="subdir", dir=temp_dir):
                task = FileSensor(
                    task_id="test",
                    filepath=os.path.join(temp_dir, "*dir"),
                    fs_conn_id="fs_default",
                    dag=self.dag,
                    timeout=0,
                )
                task._hook = self.hook

                # No files in dir
                with pytest.raises(AirflowSensorTimeout):
                    task.execute({})

    def test_wildcard_directory_with_files(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with tempfile.TemporaryDirectory(suffix="subdir", dir=temp_dir) as subdir:
                task = FileSensor(
                    task_id="test",
                    filepath=os.path.join(temp_dir, "*dir"),
                    fs_conn_id="fs_default",
                    dag=self.dag,
                    timeout=0,
                )
                task._hook = self.hook

                # `touch` the file in subdir
                open(os.path.join(subdir, "file"), "a").close()
                task.execute({})

    def test_wildcared_directory(self):
        temp_dir = tempfile.mkdtemp()
        subdir = tempfile.mkdtemp(dir=temp_dir)
        task = FileSensor(
            task_id="test",
            filepath=temp_dir + "/**",
            fs_conn_id="fs_default",
            dag=self.dag,
            timeout=0,
            poke_interval=1,
            recursive=True,
        )
        task._hook = self.hook

        try:
            # `touch` the file in subdir
            open(subdir + "/file", "a").close()
            task.execute({})
        finally:
            shutil.rmtree(temp_dir)

    def test_subdirectory_not_empty(self):
        suffix = ".txt"
        temp_dir = tempfile.mkdtemp()
        subdir = tempfile.mkdtemp(dir=temp_dir)

        with tempfile.NamedTemporaryFile(suffix=suffix, dir=subdir):
            task = FileSensor(
                task_id="test",
                filepath=temp_dir,
                fs_conn_id="fs_default",
                dag=self.dag,
                timeout=0,
            )
            task._hook = self.hook
            task.execute({})
        shutil.rmtree(temp_dir)

    def test_subdirectory_empty(self, tmp_path):
        (tmp_path / "subdir").mkdir()

        task = FileSensor(
            task_id="test",
            filepath=tmp_path.as_posix(),
            fs_conn_id="fs_default",
            dag=self.dag,
            timeout=0,
            poke_interval=1,
        )
        task._hook = self.hook

        with pytest.raises(AirflowSensorTimeout):
            task.execute({})

    def test_task_defer(self):
        task = FileSensor(
            task_id="test",
            filepath="temp_dir",
            fs_conn_id="fs_default",
            deferrable=True,
            dag=self.dag,
        )

        with pytest.raises(TaskDeferred) as exc:
            task.execute({})

        assert isinstance(exc.value.trigger, FileTrigger), "Trigger is not a FileTrigger"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import time
from unittest import mock

import pytest

from airflow.models.dag import DAG
from airflow.models.dagrun import DagRun
from airflow.providers.standard.triggers.external_task import DagStateTrigger, WorkflowTrigger
from airflow.triggers.base import TriggerEvent
from airflow.utils import timezone
from airflow.utils.state import DagRunState

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

_DATES = (
    {"run_ids": ["external_task_run_id"]}
    if AIRFLOW_V_3_0_PLUS
    else {"execution_dates": [timezone.datetime(2022, 1, 1)]}
)
key, value = next(iter(_DATES.items()))


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test only for Airflow 3")
class TestWorkflowTrigger:
    DAG_ID = "external_task"
    TASK_ID = "external_task_op"
    RUN_ID = "external_task_run_id"
    STATES = ["success", "fail"]
    LOGICAL_DATE = timezone.datetime(2022, 1, 1)

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_success(self, mock_get_count):
        """check the db count get called correctly."""
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            external_task_ids=[self.TASK_ID],
            allowed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        await fake_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert result.payload == {"status": "success"}

        mock_get_count.assert_called_once_with(
            dag_id="external_task",
            task_ids=["external_task_op"],
            logical_dates=[self.LOGICAL_DATE],
            run_ids=None,
            states=["success", "fail"],
        )
        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_failed(self, mock_get_count):
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            run_ids=[self.RUN_ID],
            external_task_ids=[self.TASK_ID],
            failed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        await fake_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "failed"}
        mock_get_count.assert_called_once_with(
            dag_id="external_task",
            task_ids=["external_task_op"],
            logical_dates=[self.LOGICAL_DATE],
            run_ids=[self.RUN_ID],
            states=["success", "fail"],
        )
        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @pytest.mark.asyncio
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    async def test_task_workflow_trigger_fail_count_eq_0(self, mock_get_count):
        mock_get_count.side_effect = [0, 1]  # First 0 for failed_states, then 1 for allowed_states

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            run_ids=[self.RUN_ID],
            external_task_ids=[self.TASK_ID],
            failed_states=self.STATES,
            allowed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        await trigger_task
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "success"}

        # Verify both calls were made
        assert mock_get_count.call_count == 2
        mock_get_count.assert_has_calls(
            [
                mock.call(
                    dag_id="external_task",
                    task_ids=["external_task_op"],
                    logical_dates=[self.LOGICAL_DATE],
                    run_ids=[self.RUN_ID],
                    states=["success", "fail"],
                ),
            ]
            * 2
        )

        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_skipped(self, mock_get_count):
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            external_task_ids=[self.TASK_ID],
            skipped_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        await fake_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "skipped"}
        mock_get_count.assert_called_once_with(
            dag_id="external_task",
            task_ids=["external_task_op"],
            logical_dates=[self.LOGICAL_DATE],
            run_ids=None,
            states=["success", "fail"],
        )

    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    @mock.patch("asyncio.sleep")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_sleep_success(self, mock_sleep, mock_get_count):
        mock_get_count.side_effect = [0, 1]

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            external_task_ids=[self.TASK_ID],
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        await trigger_task
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "success"}
        mock_get_count.assert_called()
        assert mock_get_count.call_count == 2

        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

        mock_sleep.assert_awaited()
        assert mock_sleep.await_count == 1

    @pytest.mark.parametrize(
        (
            "task_ids",
            "task_group_id",
            "states",
            "logical_dates",
            "mock_ti_count",
            "mock_task_states",
            "mock_dag_count",
            "expected",
        ),
        [
            (
                ["task_id_one", "task_id_two"],
                None,
                ["success"],
                [
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                ],
                4,
                None,
                None,
                2,
            ),
            (
                [],
                "task_group_id",
                ["success"],
                [
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                ],
                None,
                {"run_id_one": {"group1.task_id": "success"}, "run_id_two": {"group1.task_id": "success"}},
                None,
                2,
            ),
            (
                [],
                "task_group_id",
                ["success"],
                [
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                ],
                None,
                {
                    "run_id_one": {"task_group_id.task_id": "success"},
                    "run_id_two": {"task_group_id.task_id": "failed"},
                },
                None,
                1,
            ),
            (
                [],
                "task_group_id",
                ["success"],
                [
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                ],
                None,
                {
                    "run_id_one": {"task_group_id.task_id": "success"},
                    "run_id_two": {"task_group_id.task_id": "skipped"},
                },
                None,
                1,
            ),
            (
                [],
                None,
                ["success"],
                [
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                    timezone.datetime(2020, 7, 6, 13, tzinfo=timezone.utc),
                ],
                None,
                None,
                2,
                2,
            ),
        ],
        ids=[
            "with_task_ids",
            "with task_group_id only",
            "with task_group_id and some task_ids failed",
            "without task_group_id and some task_ids skipped",
            "no task_ids or task_group_id",
        ],
    )
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_ti_count")
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_task_states")
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_dr_count")
    @pytest.mark.asyncio
    async def test_get_count_af_3(
        self,
        mock_get_dr_count,
        mock_get_task_states,
        mock_get_ti_count,
        task_ids,
        task_group_id,
        states,
        logical_dates,
        mock_ti_count,
        mock_task_states,
        mock_dag_count,
        expected,
    ):
        """
        case1: when provided two task_ids, and two dag runs, the get_ti_count should return 4(each dag run returns two tasks)
                and normalized count becomes 2
        case2: when provided task_group_id, and two dag runs, the get_ti_count should return 2(each dag run returns 1 task group)
        case3: when not provided any task_ids or task_group_id, the get_dr_count should return 2(total dag runs 2)
        """

        mock_get_ti_count.return_value = mock_ti_count
        mock_get_dr_count.return_value = mock_dag_count
        mock_get_task_states.return_value = mock_task_states

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=logical_dates,
            external_task_ids=task_ids,
            external_task_group_id=task_group_id,
            allowed_states=states,
            poke_interval=0.2,
        )

        get_count_af_3 = await trigger._get_count_af_3(states)
        assert get_count_af_3 == expected

        if task_ids:
            mock_get_ti_count.assert_called_once()
            assert mock_get_ti_count.call_count == 1
            mock_get_task_states.assert_not_called()
            assert mock_get_task_states.call_count == 0
            mock_get_dr_count.assert_not_called()
            assert mock_get_dr_count.call_count == 0

        elif task_group_id:
            mock_get_task_states.assert_called_once()
            assert mock_get_task_states.call_count == 1
            mock_get_ti_count.assert_not_called()
            assert mock_get_ti_count.call_count == 0
            mock_get_dr_count.assert_not_called()
            assert mock_get_dr_count.call_count == 0

        if not task_ids and not task_group_id:
            mock_get_dr_count.assert_called_once()
            assert mock_get_dr_count.call_count == 1
            mock_get_ti_count.assert_not_called()
            assert mock_get_ti_count.call_count == 0
            mock_get_task_states.assert_not_called()
            assert mock_get_task_states.call_count == 0

    def test_serialization(self):
        """
        Asserts that the WorkflowTrigger correctly serializes its arguments and classpath.
        """
        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            logical_dates=[self.LOGICAL_DATE],
            run_ids=[self.RUN_ID],
            failed_states=["failed"],
            skipped_states=["skipped"],
            external_task_ids=[self.TASK_ID],
            allowed_states=self.STATES,
            poke_interval=5,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.standard.triggers.external_task.WorkflowTrigger"
        assert kwargs == {
            "external_dag_id": self.DAG_ID,
            "logical_dates": [self.LOGICAL_DATE],
            "run_ids": [self.RUN_ID],
            "external_task_ids": [self.TASK_ID],
            "external_task_group_id": None,
            "failed_states": ["failed"],
            "skipped_states": ["skipped"],
            "allowed_states": self.STATES,
            "poke_interval": 5,
            "soft_fail": False,
        }


@pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test only for Airflow 2")
class TestWorkflowTriggerAF2:
    DAG_ID = "external_task"
    TASK_ID = "external_task_op"
    RUN_ID = "external_task_run_id"
    STATES = ["success", "fail"]

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.providers.standard.triggers.external_task._get_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_success(self, mock_get_count):
        """check the db count get called correctly."""
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            allowed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert result.payload == {"status": "success"}

        mock_get_count.assert_called_once_with(
            dttm_filter=value,
            external_task_ids=["external_task_op"],
            external_task_group_id=None,
            external_dag_id="external_task",
            states=["success", "fail"],
        )
        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.providers.standard.triggers.external_task._get_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_failed(self, mock_get_count):
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            failed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "failed"}
        mock_get_count.assert_called_once_with(
            dttm_filter=value,
            external_task_ids=["external_task_op"],
            external_task_group_id=None,
            external_dag_id="external_task",
            states=["success", "fail"],
        )
        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @mock.patch("airflow.providers.standard.triggers.external_task._get_count")
    @mock.patch("asyncio.sleep")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_fail_count_eq_0(self, mock_sleep, mock_get_count):
        mock_get_count.side_effect = [
            0,
            1,
        ]

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            failed_states=self.STATES,
            allowed_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        await trigger_task
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "success"}

        assert mock_get_count.call_count == 2
        mock_get_count.assert_has_calls(
            [
                mock.call(
                    dttm_filter=value,
                    external_task_ids=["external_task_op"],
                    external_task_group_id=None,
                    external_dag_id="external_task",
                    states=["success", "fail"],
                ),
            ]
            * 2
        )

        mock_sleep.assert_not_called()

        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

    @pytest.mark.flaky(reruns=5)
    @mock.patch("airflow.providers.standard.triggers.external_task._get_count")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_skipped(self, mock_get_count):
        mock_get_count.side_effect = mocked_get_count

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            skipped_states=self.STATES,
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        fake_task = asyncio.create_task(fake_async_fun())
        await trigger_task
        assert fake_task.done()  # confirm that get_count is done in an async fashion
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "skipped"}
        mock_get_count.assert_called_once_with(
            dttm_filter=value,
            external_task_ids=["external_task_op"],
            external_task_group_id=None,
            external_dag_id="external_task",
            states=["success", "fail"],
        )

    @mock.patch("airflow.providers.standard.triggers.external_task._get_count")
    @mock.patch("asyncio.sleep")
    @pytest.mark.asyncio
    async def test_task_workflow_trigger_sleep_success(self, mock_sleep, mock_get_count):
        mock_get_count.side_effect = [0, 1]

        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            poke_interval=0.2,
        )

        gen = trigger.run()
        trigger_task = asyncio.create_task(gen.__anext__())
        await trigger_task
        assert trigger_task.done()
        result = trigger_task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == {"status": "success"}
        mock_get_count.assert_called()
        assert mock_get_count.call_count == 2

        # test that it returns after yielding
        with pytest.raises(StopAsyncIteration):
            await gen.__anext__()

        mock_sleep.assert_awaited()
        assert mock_sleep.await_count == 1

    def test_serialization(self):
        """
        Asserts that the WorkflowTrigger correctly serializes its arguments and classpath.
        """
        trigger = WorkflowTrigger(
            external_dag_id=self.DAG_ID,
            **_DATES,
            external_task_ids=[self.TASK_ID],
            allowed_states=self.STATES,
            poke_interval=5,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.standard.triggers.external_task.WorkflowTrigger"
        assert kwargs == {
            "external_dag_id": self.DAG_ID,
            **_DATES,
            "external_task_ids": [self.TASK_ID],
            "external_task_group_id": None,
            "failed_states": None,
            "skipped_states": None,
            "allowed_states": self.STATES,
            "poke_interval": 5,
            "soft_fail": False,
        }


class TestDagStateTrigger:
    DAG_ID = "test_dag_state_trigger"
    RUN_ID = "external_task_run_id"
    STATES = ["success", "fail"]
    EXECUTION_DATE = timezone.datetime(2022, 1, 1)

    @pytest.mark.db_test
    @pytest.mark.asyncio
    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Airflow 3 had a different implementation")
    async def test_dag_state_trigger(self, session):
        """
        Assert that the DagStateTrigger only goes off on or after a DagRun
        reaches an allowed state (i.e. SUCCESS).
        """
        dag = DAG(self.DAG_ID, schedule=None, start_date=timezone.datetime(2022, 1, 1))
        run_id_or_execution_date = (
            {"run_id": "external_task_run_id"}
            if AIRFLOW_V_3_0_PLUS
            else {"execution_date": timezone.datetime(2022, 1, 1), "run_id": "external_task_run_id"}
        )
        dag_run = DagRun(dag_id=dag.dag_id, run_type="manual", **run_id_or_execution_date)
        session.add(dag_run)
        session.commit()

        trigger = DagStateTrigger(
            dag_id=dag.dag_id,
            states=self.STATES,
            **_DATES,
            poll_interval=0.2,
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # It should not have produced a result
        assert task.done() is False

        # Progress the dag to a "success" state so that yields a TriggerEvent
        dag_run.state = DagRunState.SUCCESS
        session.commit()
        await asyncio.sleep(0.5)
        assert task.done() is True

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.db_test
    @pytest.mark.asyncio
    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Airflow 2 had a different implementation")
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_dr_count")
    async def test_dag_state_trigger_af_3(self, mock_get_dag_run_count, session):
        """
        Assert that the DagStateTrigger only goes off on or after a DagRun
        reaches an allowed state (i.e. SUCCESS).
        """

        # Mock the get_dag_run_count_by_run_ids_and_states function to return 0 first time
        mock_get_dag_run_count.return_value = 0
        dag = DAG(self.DAG_ID, schedule=None, start_date=timezone.datetime(2022, 1, 1))

        dag_run = DagRun(
            dag_id=dag.dag_id,
            run_type="manual",
            run_id="external_task_run_id",
            logical_date=timezone.datetime(2022, 1, 1),
        )
        session.add(dag_run)
        session.commit()

        trigger = DagStateTrigger(
            dag_id=dag.dag_id,
            states=self.STATES,
            run_ids=["external_task_run_id"],
            poll_interval=0.2,
            execution_dates=[timezone.datetime(2022, 1, 1)],
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # It should not have produced a result
        assert task.done() is False

        # Progress the dag to a "success" state so that yields a TriggerEvent
        dag_run.state = DagRunState.SUCCESS
        session.commit()

        # Mock the get_dag_run_count_by_run_ids_and_states function to return 1 second time
        mock_get_dag_run_count.return_value = 1
        await asyncio.sleep(0.5)
        assert task.done() is True

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.db_test
    @pytest.mark.asyncio
    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Airflow 2 had a different implementation")
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_dr_count")
    @mock.patch("airflow.sdk.execution_time.task_runner.RuntimeTaskInstance.get_dagrun_state")
    async def test_dag_state_trigger_af_3_return_type(
        self, mock_get_dagrun_state, mock_get_dag_run_count, session
    ):
        """
        Assert that the DagStateTrigger returns a tuple with classpath and event_data.
        """
        mock_get_dag_run_count.return_value = 1
        mock_get_dagrun_state.return_value = DagRunState.SUCCESS
        dag = DAG(f"{self.DAG_ID}_return_type", schedule=None, start_date=timezone.datetime(2022, 1, 1))

        dag_run = DagRun(
            dag_id=dag.dag_id,
            run_type="manual",
            run_id="external_task_run_id",
            logical_date=timezone.datetime(2022, 1, 1),
        )
        dag_run.state = DagRunState.SUCCESS
        session.add(dag_run)
        session.commit()

        trigger = DagStateTrigger(
            dag_id=dag.dag_id,
            states=self.STATES,
            run_ids=["external_task_run_id"],
            poll_interval=0.2,
            execution_dates=[timezone.datetime(2022, 1, 1)],
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)
        assert task.done() is True

        result = task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == (
            "airflow.providers.standard.triggers.external_task.DagStateTrigger",
            {
                "dag_id": "test_dag_state_trigger_return_type",
                "execution_dates": [
                    timezone.datetime(2022, 1, 1, 0, 0, tzinfo=timezone.utc),
                ],
                "external_task_run_id": DagRunState.SUCCESS,
                "poll_interval": 0.2,
                "run_ids": ["external_task_run_id"],
                "states": ["success", "fail"],
            },
        )
        asyncio.get_event_loop().stop()

    @pytest.mark.db_test
    @pytest.mark.asyncio
    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test only AF2 implementation.")
    async def test_dag_state_trigger_af_2_return_type(self, session):
        """
        Assert that the DagStateTrigger returns a tuple with classpath and event_data.
        """
        dag = DAG(f"{self.DAG_ID}_return_type", schedule=None, start_date=timezone.datetime(2022, 1, 1))

        dag_run = DagRun(
            dag_id=dag.dag_id,
            run_type="manual",
            run_id="external_task_run_id",
            execution_date=timezone.datetime(2022, 1, 1),
        )
        dag_run.state = DagRunState.SUCCESS
        session.add(dag_run)
        session.commit()

        trigger = DagStateTrigger(
            dag_id=dag.dag_id,
            states=self.STATES,
            run_ids=["external_task_run_id"],
            poll_interval=0.2,
            execution_dates=[timezone.datetime(2022, 1, 1)],
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)
        assert task.done() is True

        result = task.result()
        assert isinstance(result, TriggerEvent)
        assert result.payload == (
            "airflow.providers.standard.triggers.external_task.DagStateTrigger",
            {
                "dag_id": "test_dag_state_trigger_return_type",
                "execution_dates": [
                    timezone.datetime(2022, 1, 1, 0, 0, tzinfo=timezone.utc),
                ],
                # 'external_task_run_id': DagRunState.SUCCESS,  # This is only appended in AF3
                "poll_interval": 0.2,
                "run_ids": ["external_task_run_id"],
                "states": ["success", "fail"],
            },
        )
        asyncio.get_event_loop().stop()

    def test_serialization(self):
        """Asserts that the DagStateTrigger correctly serializes its arguments and classpath."""
        trigger = DagStateTrigger(
            dag_id=self.DAG_ID,
            states=self.STATES,
            run_ids=[TestDagStateTrigger.RUN_ID],
            execution_dates=[TestDagStateTrigger.EXECUTION_DATE],
            poll_interval=5,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.standard.triggers.external_task.DagStateTrigger"
        assert kwargs == {
            "dag_id": self.DAG_ID,
            "states": self.STATES,
            "run_ids": [TestDagStateTrigger.RUN_ID],
            "execution_dates": [TestDagStateTrigger.EXECUTION_DATE],
            "poll_interval": 5,
        }


def mocked_get_count(*args, **kwargs):
    time.sleep(0.0001)
    return 1


async def fake_async_fun():
    await asyncio.sleep(0.00005)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import subprocess
import venv
from datetime import timedelta
from importlib.util import find_spec
from subprocess import CalledProcessError

import pytest

from tests_common.test_utils.taskinstance import run_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import setup, task, teardown
else:
    from airflow.decorators import setup, task, teardown  # type: ignore[attr-defined,no-redef]

try:
    from airflow.utils import timezone  # type: ignore[attr-defined]
except AttributeError:
    from airflow.sdk import timezone

pytestmark = pytest.mark.db_test


DEFAULT_DATE = timezone.datetime(2016, 1, 1)
END_DATE = timezone.datetime(2016, 1, 2)
INTERVAL = timedelta(hours=12)
FROZEN_NOW = timezone.datetime(2016, 1, 2, 12, 1, 1)
DILL_INSTALLED = find_spec("dill") is not None
DILL_MARKER = pytest.mark.skipif(not DILL_INSTALLED, reason="`dill` is not installed")
CLOUDPICKLE_INSTALLED = find_spec("cloudpickle") is not None
CLOUDPICKLE_MARKER = pytest.mark.skipif(not CLOUDPICKLE_INSTALLED, reason="`cloudpickle` is not installed")

TI_CONTEXT_ENV_VARS = [
    "AIRFLOW_CTX_DAG_ID",
    "AIRFLOW_CTX_TASK_ID",
    "AIRFLOW_CTX_LOGICAL_DATE",
    "AIRFLOW_CTX_DAG_RUN_ID",
]


@pytest.fixture(scope="module")
def venv_python(tmp_path_factory):
    venv_dir = tmp_path_factory.mktemp("venv")
    venv.create(venv_dir, with_pip=False)
    return (venv_dir / "bin" / "python").resolve(strict=True).as_posix()


@pytest.fixture(scope="module")
def venv_python_with_cloudpickle_and_dill(tmp_path_factory):
    venv_dir = tmp_path_factory.mktemp("venv_serializers")
    venv.create(venv_dir, with_pip=True)
    python_path = (venv_dir / "bin" / "python").resolve(strict=True).as_posix()
    subprocess.check_call([python_path, "-m", "pip", "install", "cloudpickle", "dill"])
    return python_path


class TestExternalPythonDecorator:
    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
        ],
    )
    def test_with_serializer_works(self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill):
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f():
            """Import cloudpickle/dill to double-check it is installed ."""
            import cloudpickle  # noqa: F401
            import dill  # noqa: F401

        with dag_maker(serialized=True):
            v = f()

        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
        ],
    )
    def test_with_templated_python_serializer(
        self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill
    ):
        # add template that produces empty string when rendered
        templated_python_with_cloudpickle = venv_python_with_cloudpickle_and_dill + "{{ '' }}"

        @task.external_python(python=templated_python_with_cloudpickle, serializer=serializer)
        def f():
            """Import cloudpickle/dill to double-check it is installed ."""
            import cloudpickle  # noqa: F401
            import dill  # noqa: F401

        with dag_maker(serialized=True):
            v = f()
        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
        ],
    )
    def test_no_advanced_serializer_installed(self, serializer, dag_maker, venv_python):
        @task.external_python(python=venv_python, serializer=serializer)
        def f():
            pass

        with dag_maker(serialized=True):
            v = f()
        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]

        with pytest.raises(CalledProcessError):
            run_task_instance(ti, v.operator)

    def test_exception_raises_error(self, dag_maker, venv_python):
        @task.external_python(python=venv_python)
        def f():
            raise Exception

        with dag_maker(serialized=True):
            v = f()
        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]

        with pytest.raises(CalledProcessError):
            run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    def test_with_args(self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill):
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f(a, b, c=False, d=False):
            if a == 0 and b == 1 and c and not d:
                return True
            raise Exception

        with dag_maker(serialized=True):
            v = f(0, 1, c=True)
        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    def test_return_none(self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill):
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f():
            return None

        with dag_maker(serialized=True):
            v = f()

        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    def test_nonimported_as_arg(self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill):
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f(_):
            return None

        with dag_maker(serialized=True):
            v = f(datetime.datetime.now(tz=datetime.timezone.utc))

        dr = dag_maker.create_dagrun()
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    def test_marking_external_python_task_as_setup(
        self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill
    ):
        @setup
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f():
            return 1

        with dag_maker(serialized=True) as dag:
            v = f()
        dr = dag_maker.create_dagrun()

        assert len(dag.task_group.children) == 1
        setup_task = dag.task_group.children["f"]
        assert setup_task.is_setup
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    def test_marking_external_python_task_as_teardown(
        self, serializer, dag_maker, venv_python_with_cloudpickle_and_dill
    ):
        @teardown
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f():
            return 1

        with dag_maker(serialized=True) as dag:
            v = f()
        dr = dag_maker.create_dagrun()

        assert len(dag.task_group.children) == 1
        teardown_task = dag.task_group.children["f"]
        assert teardown_task.is_teardown
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)

    @pytest.mark.parametrize(
        "serializer",
        [
            pytest.param("pickle", id="pickle"),
            pytest.param("dill", marks=DILL_MARKER, id="dill"),
            pytest.param("cloudpickle", marks=CLOUDPICKLE_MARKER, id="cloudpickle"),
            pytest.param(None, id="default"),
        ],
    )
    @pytest.mark.parametrize("on_failure_fail_dagrun", [True, False])
    def test_marking_external_python_task_as_teardown_with_on_failure_fail(
        self, serializer, dag_maker, on_failure_fail_dagrun, venv_python_with_cloudpickle_and_dill
    ):
        @teardown(on_failure_fail_dagrun=on_failure_fail_dagrun)
        @task.external_python(python=venv_python_with_cloudpickle_and_dill, serializer=serializer)
        def f():
            return 1

        with dag_maker(serialized=True) as dag:
            v = f()
        dr = dag_maker.create_dagrun()

        assert len(dag.task_group.children) == 1
        teardown_task = dag.task_group.children["f"]
        assert teardown_task.is_teardown
        assert teardown_task.on_failure_fail_dagrun is on_failure_fail_dagrun
        ti = dr.get_task_instances()[0]
        run_task_instance(ti, v.operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Hook for additional Package Indexes (Python)."""

from __future__ import annotations

import subprocess
from typing import Any
from urllib.parse import quote, urlparse

from airflow.providers.common.compat.sdk import BaseHook


class PackageIndexHook(BaseHook):
    """Specify package indexes/Python package sources using Airflow connections."""

    conn_name_attr = "pi_conn_id"
    default_conn_name = "package_index_default"
    conn_type = "package_index"
    hook_name = "Package Index (Python)"

    def __init__(self, pi_conn_id: str = default_conn_name, **kwargs) -> None:
        super().__init__(**kwargs)
        self.pi_conn_id = pi_conn_id
        self.conn = None

    @staticmethod
    def get_ui_field_behaviour() -> dict[str, Any]:
        """Return custom field behaviour."""
        return {
            "hidden_fields": ["schema", "port", "extra"],
            "relabeling": {"host": "Package Index URL"},
            "placeholders": {
                "host": "Example: https://my-package-mirror.net/pypi/repo-name/simple",
                "login": "Username for package index",
                "password": "Password for package index (will be masked)",
            },
        }

    @staticmethod
    def _get_basic_auth_conn_url(index_url: str, user: str | None, password: str | None) -> str:
        """Return a connection URL with basic auth credentials based on connection config."""
        url = urlparse(index_url)
        host = url.netloc.split("@")[-1]
        if user:
            if password:
                host = f"{quote(user)}:{quote(password)}@{host}"
            else:
                host = f"{quote(user)}@{host}"
        return url._replace(netloc=host).geturl()

    def get_conn(self) -> Any:
        """Return connection for the hook."""
        return self.get_connection_url()

    def get_connection_url(self) -> Any:
        """Return a connection URL with embedded credentials."""
        conn = self.get_connection(self.pi_conn_id)
        index_url = conn.host
        if not index_url:
            raise ValueError("Please provide an index URL.")
        return self._get_basic_auth_conn_url(index_url, conn.login, conn.password)

    def test_connection(self) -> tuple[bool, str]:
        """Test connection to package index url."""
        conn_url = self.get_connection_url()
        proc = subprocess.run(
            ["pip", "search", "not-existing-test-package", "--no-input", "--index", conn_url],
            check=False,
            capture_output=True,
        )
        conn = self.get_connection(self.pi_conn_id)
        if proc.returncode not in [
            0,  # executed successfully, found package
            23,  # executed successfully, didn't find any packages
            #      (but we do not expect it to find 'not-existing-test-package')
        ]:
            return False, f"Connection test to {conn.host} failed. Error: {str(proc.stderr)}"

        return True, f"Connection to {conn.host} tested successfully!"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime

import pytest
from sqlalchemy import select

from airflow.models.taskinstance import TaskInstance as TI
from airflow.providers.standard.operators.branch import BaseBranchOperator
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.utils.skipmixin import XCOM_SKIPMIXIN_FOLLOWED, XCOM_SKIPMIXIN_KEY
from airflow.timetables.base import DataInterval
from airflow.utils import timezone
from airflow.utils.state import State
from airflow.utils.types import DagRunType

try:
    from airflow.sdk.definitions.taskgroup import TaskGroup
except ImportError:
    # Fallback for Airflow < 3.1
    from airflow.utils.task_group import TaskGroup  # type: ignore[no-redef]

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_1, AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.providers.common.compat.sdk import DownstreamTasksSkipped
    from airflow.utils.types import DagRunTriggeredByType

pytestmark = pytest.mark.db_test

DEFAULT_DATE = timezone.datetime(2016, 1, 1)
INTERVAL = datetime.timedelta(hours=12)


class ChooseBranchOne(BaseBranchOperator):
    def choose_branch(self, context):
        return "branch_1"


class ChooseBranchOneTwo(BaseBranchOperator):
    def choose_branch(self, context):
        return ["branch_1", "branch_2"]


class ChooseBranchThree(BaseBranchOperator):
    def choose_branch(self, context):
        return ["branch_3"]


class TestBranchOperator:
    def test_without_dag_run(self, dag_maker):
        """This checks the defensive against non-existent tasks in a dag run"""
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
        dr = dag_maker.create_dagrun(**triggered_by_kwargs)

        if AIRFLOW_V_3_0_1:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                dag_maker.run_ti("make_choice", dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            dag_maker.run_ti("make_choice", dr)

            ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

            for ti in dag_maker.session.scalars(
                select(TI).where(TI.dag_id == dag_id, ti_date == DEFAULT_DATE)
            ):
                if ti.task_id == "make_choice":
                    assert ti.state == State.SUCCESS
                elif ti.task_id == "branch_1":
                    assert ti.state == State.NONE
                elif ti.task_id == "branch_2":
                    assert ti.state == State.SKIPPED
                else:
                    raise Exception

    def test_branch_list_without_dag_run(self, dag_maker):
        """This checks if the BranchOperator supports branching off to a list of tasks."""
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_3 = EmptyOperator(task_id="branch_3")
            branch_op = ChooseBranchOneTwo(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
            branch_3.set_upstream(branch_op)
        dr = dag_maker.create_dagrun(**triggered_by_kwargs)

        if AIRFLOW_V_3_0_1:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                dag_maker.run_ti("make_choice", dr)

            assert exc_info.value.tasks == [("branch_3", -1)]
        else:
            dag_maker.run_ti("make_choice", dr)

            expected = {
                "make_choice": State.SUCCESS,
                "branch_1": State.NONE,
                "branch_2": State.NONE,
                "branch_3": State.SKIPPED,
            }

            ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

            for ti in dag_maker.session.scalars(
                select(TI).where(TI.dag_id == dag_id, ti_date == DEFAULT_DATE)
            ):
                if ti.task_id in expected:
                    assert ti.state == expected[ti.task_id]
                else:
                    raise Exception

    def test_with_dag_run(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
        if AIRFLOW_V_3_0_1:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                dag_maker.run_ti("make_choice", dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            dag_maker.run_ti("make_choice", dr)

            expected = {
                "make_choice": State.SUCCESS,
                "branch_1": State.NONE,
                "branch_2": State.SKIPPED,
            }

            ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

            for ti in dag_maker.session.scalars(
                select(TI).where(TI.dag_id == dag_id, ti_date == DEFAULT_DATE)
            ):
                if ti.task_id in expected:
                    assert ti.state == expected[ti.task_id]
                else:
                    raise Exception

    def test_with_skip_in_branch_downstream_dependencies(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_op >> branch_1 >> branch_2
            branch_op >> branch_2

        if AIRFLOW_V_3_0_PLUS:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )
        else:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

        dag_maker.run_ti("make_choice", dr)

        expected = {
            "make_choice": State.SUCCESS,
            "branch_1": State.NONE,
            "branch_2": State.NONE,
        }

        ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

        for ti in dag_maker.session.scalars(select(TI).where(TI.dag_id == dag_id, ti_date == DEFAULT_DATE)):
            if ti.task_id in expected:
                assert ti.state == expected[ti.task_id]
            else:
                raise Exception

    def test_xcom_push(self, dag_maker):
        dag_id = "branch_operator_test"

        triggered_by_kwargs = (
            {
                "triggered_by": DagRunTriggeredByType.TEST,
                "logical_date": DEFAULT_DATE,
            }
            if AIRFLOW_V_3_0_PLUS
            else {"execution_date": DEFAULT_DATE}
        )
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)

        dr = dag_maker.create_dagrun(
            run_type=DagRunType.MANUAL,
            start_date=timezone.utcnow(),
            state=State.RUNNING,
            data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
            **triggered_by_kwargs,
        )

        if AIRFLOW_V_3_0_1:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                dag_maker.run_ti("make_choice", dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            dag_maker.run_ti("make_choice", dr)
            branch_op_ti = dr.get_task_instance("make_choice")
            assert branch_op_ti.xcom_pull(task_ids="make_choice", key=XCOM_SKIPMIXIN_KEY) == {
                XCOM_SKIPMIXIN_FOLLOWED: ["branch_1"]
            }

    def test_with_dag_run_task_groups(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_3 = TaskGroup("branch_3")
            EmptyOperator(task_id="task_1", task_group=branch_3)
            EmptyOperator(task_id="task_2", task_group=branch_3)
            branch_op = ChooseBranchThree(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
            branch_3.set_upstream(branch_op)

        if AIRFLOW_V_3_0_1:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                dag_maker.run_ti("make_choice", dr)

            assert set(exc_info.value.tasks) == {("branch_1", -1), ("branch_2", -1)}
        else:
            dr = dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )
            dag_maker.run_ti("make_choice", dr)

            ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

            for ti in dag_maker.session.scalars(
                select(TI).where(TI.dag_id == dag_id, ti_date == DEFAULT_DATE)
            ):
                if ti.task_id == "make_choice":
                    assert ti.state == State.SUCCESS
                elif ti.task_id == "branch_1":
                    assert ti.state == State.SKIPPED
                elif ti.task_id == "branch_2":
                    assert ti.state == State.SKIPPED
                elif ti.task_id == "branch_3.task_1":
                    assert ti.state == State.NONE
                elif ti.task_id == "branch_3.task_2":
                    assert ti.state == State.NONE
                else:
                    raise Exception



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from airflow.providers.common.compat.sdk import task_decorator_factory
from airflow.providers.standard.decorators.python import _PythonDecoratedOperator
from airflow.providers.standard.operators.python import BranchPythonVirtualenvOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import TaskDecorator


class _BranchPythonVirtualenvDecoratedOperator(_PythonDecoratedOperator, BranchPythonVirtualenvOperator):
    """Wraps a Python callable and captures args/kwargs when called for execution."""

    template_fields = BranchPythonVirtualenvOperator.template_fields
    custom_operator_name: str = "@task.branch_virtualenv"


def branch_virtualenv_task(
    python_callable: Callable | None = None, multiple_outputs: bool | None = None, **kwargs
) -> TaskDecorator:
    """
    Wrap a python function into a BranchPythonVirtualenvOperator.

    For more information on how to use this operator, take a look at the guide:
    :ref:`concepts:branching`

    Accepts kwargs for operator kwarg. Can be reused in a single DAG.

    :param python_callable: Function to decorate
    :param multiple_outputs: if set, function return value will be
        unrolled to multiple XCom values. Dict will unroll to xcom values with keys as XCom keys.
        Defaults to False.
    """
    return task_decorator_factory(
        python_callable=python_callable,
        multiple_outputs=multiple_outputs,
        decorated_operator_class=_BranchPythonVirtualenvDecoratedOperator,
        **kwargs,
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING
from unittest import mock

import pendulum
import pytest

from airflow.models import DagBag, TaskInstance
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.utils.sensor_helper import (
    _count_stmt,
    _get_count,
    _get_external_task_group_task_ids,
)
from airflow.utils.state import DagRunState, TaskInstanceState
from airflow.utils.types import DagRunType

from tests_common.test_utils import db
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

try:
    from airflow.sdk.definitions.taskgroup import TaskGroup
except ImportError:  # Fallback for Airflow < 3.1
    from airflow.utils.task_group import TaskGroup  # type: ignore[no-redef]
try:
    from airflow.sdk import timezone
except ImportError:  # Fallback for Airflow < 3.1
    from airflow.utils import timezone  # type: ignore[attr-defined,no-redef]

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session

    from airflow.serialization.definitions.dag import SerializedDAG

TI = TaskInstance


@pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test only for AF2")
@pytest.mark.db_test
class TestSensorHelper:
    DAG_ID = "test_dag"
    TASK_ID = "test_task"
    TASK_ID_LIST = ["task_0", "task_1"]
    TASK_GROUP_ID = "test_task_group"
    TASK_GROUP = {TASK_GROUP_ID: ["subtask_0", "subtask_1"]}
    DTTM_FILTER = [
        datetime.datetime(2025, 1, 1),
        datetime.datetime(2025, 1, 2),
        datetime.datetime(2025, 1, 3),
    ]
    ALLOWED_STATES = [TaskInstanceState.SUCCESS.value, TaskInstanceState.SKIPPED.value]

    @pytest.fixture(autouse=True)
    def setup_test_cases(self, session):
        self._clean_db()
        yield
        self._clean_db()

    @staticmethod
    def _clean_db():
        db.clear_db_dags()
        db.clear_db_jobs()
        db.clear_db_runs()

    @staticmethod
    def create_dag_run(
        dag: SerializedDAG,
        *,
        task_states: Mapping[str, TaskInstanceState] | None = None,
        execution_date: datetime.datetime | None = None,
        session: Session,
    ):
        now = timezone.utcnow()
        execution_date = pendulum.instance(execution_date or now)
        run_type = DagRunType.MANUAL
        data_interval = dag.timetable.infer_manual_data_interval(run_after=execution_date)
        dag_run = dag.create_dagrun(
            run_id=dag.timetable.generate_run_id(
                run_type=run_type,
                run_after=execution_date,
                data_interval=None,
                logical_date=execution_date,
            ),
            run_type=run_type,
            data_interval=data_interval,
            start_date=now,
            state=DagRunState.SUCCESS,  # not important
            external_trigger=False,
            execution_date=execution_date,  # type: ignore[call-arg]
        )

        if task_states is not None:
            for task_id, task_state in task_states.items():
                ti = dag_run.get_task_instance(task_id)
                if TYPE_CHECKING:
                    assert ti
                ti.set_state(task_state, session)
            session.flush()

    @pytest.mark.parametrize(
        "dttm_to_task_state",
        [
            {
                datetime.datetime(2025, 1, 1): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SUCCESS.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.SKIPPED.value,
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "task_0": TaskInstanceState.SKIPPED.value,
                    "task_1": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_0": TaskInstanceState.SKIPPED.value,
                    "test_task_group.subtask_1": TaskInstanceState.SUCCESS.value,
                },
            },
            {
                datetime.datetime(2025, 1, 1): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.FAILED.value,
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "task_0": TaskInstanceState.SKIPPED.value,
                    "task_1": TaskInstanceState.FAILED.value,
                    "test_task_group.subtask_0": TaskInstanceState.FAILED.value,
                    "test_task_group.subtask_1": TaskInstanceState.SUCCESS.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "task_0": TaskInstanceState.FAILED.value,
                    "task_1": TaskInstanceState.FAILED.value,
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
            },
        ],  # these can be any TaskInstanceState
    )
    @pytest.mark.need_serialized_dag
    def test_count_stmt(self, dttm_to_task_state, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID, session=session) as dag:
            for task_id in self.TASK_ID_LIST:
                EmptyOperator(task_id=task_id)

            with TaskGroup(group_id=self.TASK_GROUP_ID):
                for subtask_id in self.TASK_GROUP[self.TASK_GROUP_ID]:
                    EmptyOperator(task_id=subtask_id)

        for dttm in self.DTTM_FILTER:
            self.create_dag_run(
                dag=dag, task_states=dttm_to_task_state[dttm], execution_date=dttm, session=session
            )

        dttm_filter = [pendulum.instance(dttm) for dttm in self.DTTM_FILTER]
        count = session.scalar(
            _count_stmt(
                model=TI, states=self.ALLOWED_STATES, dttm_filter=dttm_filter, external_dag_id=self.DAG_ID
            )
        )

        allowed_task_instance_states = [
            state
            for task_states in dttm_to_task_state.values()
            for state in task_states.values()
            if state in self.ALLOWED_STATES
        ]

        allowed_state_count = len(allowed_task_instance_states)
        assert count == allowed_state_count

    @pytest.mark.need_serialized_dag
    def test_get_external_task_group_task_ids(self, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID) as dag:
            with TaskGroup(group_id=self.TASK_GROUP_ID):
                for subtask_id in self.TASK_GROUP[self.TASK_GROUP_ID]:
                    EmptyOperator(task_id=subtask_id)

        for dttm in self.DTTM_FILTER:
            self.create_dag_run(dag=dag, execution_date=dttm, session=session)

        with mock.patch.object(DagBag, "get_dag") as mock_get_dag:
            mock_get_dag.return_value = dag

            dttm_filter = [pendulum.instance(dttm) for dttm in self.DTTM_FILTER]
            external_task_group_task_ids = _get_external_task_group_task_ids(
                dttm_filter=dttm_filter,
                external_task_group_id=self.TASK_GROUP_ID,
                external_dag_id=self.DAG_ID,
                session=session,
            )

        assert len(external_task_group_task_ids) == 6
        assert external_task_group_task_ids.count(("test_task_group.subtask_0", -1)) == 3
        assert external_task_group_task_ids.count(("test_task_group.subtask_1", -1)) == 3

    @pytest.mark.parametrize(
        "state",
        [
            TaskInstanceState.SUCCESS.value,
            TaskInstanceState.SKIPPED.value,
            TaskInstanceState.FAILED.value,
        ],
    )
    @pytest.mark.need_serialized_dag
    def test_get_count_with_different_states(self, state, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID) as dag:
            EmptyOperator(task_id=self.TASK_ID)

        now = timezone.utcnow()
        self.create_dag_run(dag, task_states={self.TASK_ID: state}, execution_date=now, session=session)

        count = _get_count(
            dttm_filter=[now],
            external_task_ids=[self.TASK_ID],
            external_task_group_id=None,
            external_dag_id=self.DAG_ID,
            states=[state],
            session=session,
        )

        assert count == 1

    @pytest.mark.parametrize(
        "task_states",
        [
            {
                datetime.datetime(2025, 1, 1): TaskInstanceState.SUCCESS.value,
                datetime.datetime(2025, 1, 2): TaskInstanceState.SKIPPED.value,
                datetime.datetime(2025, 1, 3): TaskInstanceState.FAILED.value,
            },
            {
                datetime.datetime(2025, 1, 1): TaskInstanceState.FAILED.value,
                datetime.datetime(2025, 1, 2): TaskInstanceState.FAILED.value,
                datetime.datetime(2025, 1, 3): TaskInstanceState.SUCCESS.value,
            },
        ],  # these can be any TaskInstanceState
    )
    @pytest.mark.need_serialized_dag
    def test_get_count_with_one_task(self, task_states, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID) as dag:
            EmptyOperator(task_id=self.TASK_ID)

        for dttm in self.DTTM_FILTER:
            self.create_dag_run(
                dag=dag, task_states={self.TASK_ID: task_states[dttm]}, execution_date=dttm, session=session
            )

        dttm_filter = [pendulum.instance(dttm) for dttm in self.DTTM_FILTER]
        count = _get_count(
            dttm_filter=dttm_filter,
            external_task_ids=[self.TASK_ID],
            external_task_group_id=None,
            external_dag_id=self.DAG_ID,
            states=self.ALLOWED_STATES,
            session=session,
        )

        allowed_state_count = len([state for state in task_states.values() if state in self.ALLOWED_STATES])
        assert count == allowed_state_count

    @pytest.mark.parametrize(
        "dttm_to_task_state",
        [
            {
                datetime.datetime(2025, 1, 1): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.SUCCESS.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "task_0": TaskInstanceState.SKIPPED.value,
                    "task_1": TaskInstanceState.SUCCESS.value,
                },
            },
            {
                datetime.datetime(2025, 1, 1): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.FAILED.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "task_0": TaskInstanceState.FAILED.value,
                    "task_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "task_0": TaskInstanceState.SUCCESS.value,
                    "task_1": TaskInstanceState.SKIPPED.value,
                },
            },
        ],  # these can be any TaskInstanceState
    )
    @pytest.mark.need_serialized_dag
    def test_get_count_with_multiple_tasks(self, dttm_to_task_state, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID) as dag:
            for task_id in self.TASK_ID_LIST:
                EmptyOperator(task_id=task_id)

        for dttm in self.DTTM_FILTER:
            self.create_dag_run(
                dag=dag, task_states=dttm_to_task_state[dttm], execution_date=dttm, session=session
            )

        dttm_filter = [pendulum.instance(dttm) for dttm in self.DTTM_FILTER]
        count = _get_count(
            dttm_filter=dttm_filter,
            external_task_ids=self.TASK_ID_LIST,
            external_task_group_id=None,
            external_dag_id=self.DAG_ID,
            states=self.ALLOWED_STATES,
            session=session,
        )

        allowed_task_instance_states = [
            state
            for task_states in dttm_to_task_state.values()
            for state in task_states.values()
            if state in self.ALLOWED_STATES
        ]

        allowed_state_count = len(allowed_task_instance_states) / len(self.TASK_ID_LIST)
        assert count == allowed_state_count

    @pytest.mark.parametrize(
        "dttm_to_subtask_state",
        [
            {
                datetime.datetime(2025, 1, 1): {
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SUCCESS.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "test_task_group.subtask_0": TaskInstanceState.SKIPPED.value,
                    "test_task_group.subtask_1": TaskInstanceState.SUCCESS.value,
                },
            },
            {
                datetime.datetime(2025, 1, 1): {
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.FAILED.value,
                },
                datetime.datetime(2025, 1, 2): {
                    "test_task_group.subtask_0": TaskInstanceState.FAILED.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
                datetime.datetime(2025, 1, 3): {
                    "test_task_group.subtask_0": TaskInstanceState.SUCCESS.value,
                    "test_task_group.subtask_1": TaskInstanceState.SKIPPED.value,
                },
            },
        ],  # these can be any TaskInstanceState
    )
    @pytest.mark.need_serialized_dag
    def test_get_count_with_task_group(self, dttm_to_subtask_state, dag_maker, session):
        with dag_maker(dag_id=self.DAG_ID, session=session) as dag:
            with TaskGroup(group_id=self.TASK_GROUP_ID):
                for subtask_id in self.TASK_GROUP[self.TASK_GROUP_ID]:
                    EmptyOperator(task_id=subtask_id)

        for dttm in self.DTTM_FILTER:
            self.create_dag_run(
                dag=dag, task_states=dttm_to_subtask_state[dttm], execution_date=dttm, session=session
            )

        with mock.patch.object(DagBag, "get_dag") as mock_get_dag:
            mock_get_dag.return_value = dag

            dttm_filter = [pendulum.instance(dttm) for dttm in self.DTTM_FILTER]
            count = _get_count(
                dttm_filter=dttm_filter,
                external_task_ids=None,
                external_task_group_id=self.TASK_GROUP_ID,
                external_dag_id=self.DAG_ID,
                states=self.ALLOWED_STATES,
                session=session,
            )

        allowed_task_instance_states = [
            state
            for task_states in dttm_to_subtask_state.values()
            for state in task_states.values()
            if state in self.ALLOWED_STATES
        ]

        allowed_state_count = len(allowed_task_instance_states) / len(self.TASK_ID_LIST)
        assert count == allowed_state_count


@pytest.mark.parametrize(
    ("run_id_task_state_map", "states", "expected_count"),
    [
        pytest.param(
            {
                "run_id_1": {"task_id_1": "success", "task_id_2": "success"},
            },
            ["success"],
            1,
            id="single_runid_with_success",
        ),
        pytest.param(
            {
                "run_id_1": {"task_id_1": "success", "task_id_2": "success"},
                "run_id_2": {"task_id_1": "success", "task_id_2": "success"},
            },
            ["success"],
            2,
            id="multiple_runid_with_success",
        ),
        pytest.param(
            {
                "run_id_1": {"task_id_1": "failed", "task_id_2": "failed"},
            },
            ["failed"],
            1,
            id="single_runid_with_failed",
        ),
        pytest.param(
            {
                "run_id_1": {"task_id_1": "success", "task_id_2": "failed"},
            },
            ["failed"],
            0,
            id="single_runid_with_no_success",
        ),
    ],
)
@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test only for AF3")
def test_get_count_by_matched_states(
    run_id_task_state_map,
    states,
    expected_count,
):
    from airflow.providers.standard.utils.sensor_helper import _get_count_by_matched_states

    count = _get_count_by_matched_states(run_id_task_state_map=run_id_task_state_map, states=states)
    assert count == expected_count



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example DAG demonstrating the usage of DateTimeBranchOperator with datetime as well as time objects as
targets.
"""

from __future__ import annotations

import pendulum

from airflow.providers.standard.operators.datetime import BranchDateTimeOperator
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.sdk import DAG

dag1 = DAG(
    dag_id="example_branch_datetime_operator",
    start_date=pendulum.datetime(2021, 1, 1, tz="UTC"),
    catchup=False,
    tags=["example"],
    schedule="@daily",
)

# [START howto_branch_datetime_operator]
empty_task_11 = EmptyOperator(task_id="date_in_range", dag=dag1)
empty_task_21 = EmptyOperator(task_id="date_outside_range", dag=dag1)

cond1 = BranchDateTimeOperator(
    task_id="datetime_branch",
    follow_task_ids_if_true=["date_in_range"],
    follow_task_ids_if_false=["date_outside_range"],
    target_upper=pendulum.datetime(2020, 10, 10, 15, 0, 0),
    target_lower=pendulum.datetime(2020, 10, 10, 14, 0, 0),
    dag=dag1,
)

# Run empty_task_11 if cond1 executes between 2020-10-10 14:00:00 and 2020-10-10 15:00:00
cond1 >> [empty_task_11, empty_task_21]
# [END howto_branch_datetime_operator]


dag2 = DAG(
    dag_id="example_branch_datetime_operator_2",
    start_date=pendulum.datetime(2021, 1, 1, tz="UTC"),
    catchup=False,
    tags=["example"],
    schedule="@daily",
)
# [START howto_branch_datetime_operator_next_day]
empty_task_12 = EmptyOperator(task_id="date_in_range", dag=dag2)
empty_task_22 = EmptyOperator(task_id="date_outside_range", dag=dag2)

cond2 = BranchDateTimeOperator(
    task_id="datetime_branch",
    follow_task_ids_if_true=["date_in_range"],
    follow_task_ids_if_false=["date_outside_range"],
    target_upper=pendulum.time(0, 0, 0),
    target_lower=pendulum.time(15, 0, 0),
    dag=dag2,
)

# Since target_lower happens after target_upper, target_upper will be moved to the following day
# Run empty_task_12 if cond2 executes between 15:00:00, and 00:00:00 of the following day
cond2 >> [empty_task_12, empty_task_22]
# [END howto_branch_datetime_operator_next_day]

dag3 = DAG(
    dag_id="example_branch_datetime_operator_3",
    start_date=pendulum.datetime(2021, 1, 1, tz="UTC"),
    catchup=False,
    tags=["example"],
    schedule="@daily",
)
# [START howto_branch_datetime_operator_logical_date]
empty_task_13 = EmptyOperator(task_id="date_in_range", dag=dag3)
empty_task_23 = EmptyOperator(task_id="date_outside_range", dag=dag3)

cond3 = BranchDateTimeOperator(
    task_id="datetime_branch",
    use_task_logical_date=True,
    follow_task_ids_if_true=["date_in_range"],
    follow_task_ids_if_false=["date_outside_range"],
    target_upper=pendulum.datetime(2020, 10, 10, 15, 0, 0),
    target_lower=pendulum.datetime(2020, 10, 10, 14, 0, 0),
    dag=dag3,
)

# Run empty_task_13 if cond3 executes between 2020-10-10 14:00:00 and 2020-10-10 15:00:00
cond3 >> [empty_task_13, empty_task_23]
# [END howto_branch_datetime_operator_logical_date]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime

import pytest
import time_machine
from sqlalchemy import delete

from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance as TI
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.standard.operators.datetime import BranchDateTimeOperator
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.timetables.base import DataInterval
from airflow.utils.session import create_session
from airflow.utils.state import State

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_1, AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_1_PLUS

if AIRFLOW_V_3_1_PLUS:
    from airflow.sdk import timezone
else:
    from airflow.utils import timezone  # type: ignore[attr-defined,no-redef]

pytestmark = pytest.mark.db_test

DEFAULT_DATE = timezone.datetime(2016, 1, 1)
INTERVAL = datetime.timedelta(hours=12)


class TestBranchDateTimeOperator:
    @classmethod
    def setup_class(cls):
        with create_session() as session:
            session.execute(delete(DagRun))
            session.execute(delete(TI))

    targets = [
        (datetime.datetime(2020, 7, 7, 10, 0, 0), datetime.datetime(2020, 7, 7, 11, 0, 0)),
        (datetime.time(10, 0, 0), datetime.time(11, 0, 0)),
        (datetime.datetime(2020, 7, 7, 10, 0, 0), datetime.time(11, 0, 0)),
        (datetime.time(10, 0, 0), datetime.datetime(2020, 7, 7, 11, 0, 0)),
    ]

    @pytest.fixture(autouse=True)
    def base_tests_setup(self, dag_maker):
        self.dag_maker = dag_maker  # Store dag_maker for use in test methods
        with dag_maker(
            "branch_datetime_operator_test",
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ) as dag:
            self.dag = dag
            self.branch_1 = EmptyOperator(task_id="branch_1")
            self.branch_2 = EmptyOperator(task_id="branch_2")

            self.branch_op = BranchDateTimeOperator(
                task_id="datetime_branch",
                follow_task_ids_if_true="branch_1",
                follow_task_ids_if_false="branch_2",
                target_upper=datetime.datetime(2020, 7, 7, 11, 0, 0),
                target_lower=datetime.datetime(2020, 7, 7, 10, 0, 0),
            )

            self.branch_1.set_upstream(self.branch_op)
            self.branch_2.set_upstream(self.branch_op)

        self.dr = dag_maker.create_dagrun(
            run_id="manual__",
            start_date=DEFAULT_DATE,
            logical_date=DEFAULT_DATE,
            state=State.RUNNING,
            data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
        )

    def teardown_method(self):
        with create_session() as session:
            session.execute(delete(DagRun))
            session.execute(delete(TI))

    def _assert_task_ids_match_states(self, task_ids_to_states):
        """Helper that asserts task instances with a given id are in a given state"""
        tis = self.dr.get_task_instances()
        for ti in tis:
            try:
                expected_state = task_ids_to_states[ti.task_id]
            except KeyError:
                raise ValueError(f"Invalid task id {ti.task_id} found!")
            else:
                assert ti.state == (expected_state), (
                    f"Task {ti.task_id} has state {ti.state} instead of expected {expected_state}"
                )

    def test_no_target_time(self):
        """Check if BranchDateTimeOperator raises exception on missing target"""
        with pytest.raises(AirflowException):
            BranchDateTimeOperator(
                task_id="datetime_branch",
                follow_task_ids_if_true="branch_1",
                follow_task_ids_if_false="branch_2",
                target_upper=None,
                target_lower=None,
                dag=self.dag_maker.dag,
            )

    @pytest.mark.parametrize(
        ("target_lower", "target_upper"),
        targets,
    )
    @time_machine.travel("2020-07-07 10:54:05")
    def test_branch_datetime_operator_falls_within_range(self, target_lower, target_upper):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_lower = target_lower
        self.branch_op.target_upper = target_upper
        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)
            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.NONE,
                    "branch_2": State.SKIPPED,
                }
            )

    @pytest.mark.parametrize(
        ("target_lower", "target_upper"),
        targets,
    )
    @pytest.mark.parametrize(
        "date",
        (
            datetime.datetime(2020, 7, 7, 12, 0, 0, tzinfo=datetime.timezone.utc),
            datetime.datetime(2020, 6, 7, 9, 0, 0, tzinfo=datetime.timezone.utc),
        ),
    )
    def test_branch_datetime_operator_falls_outside_range(self, date, target_lower, target_upper):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_lower = target_lower
        self.branch_op.target_upper = target_upper

        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info, time_machine.travel(date):
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_1", -1)]
        else:
            with time_machine.travel(date):
                self.dag_maker.run_ti("datetime_branch", self.dr)

                self._assert_task_ids_match_states(
                    {
                        "datetime_branch": State.SUCCESS,
                        "branch_1": State.SKIPPED,
                        "branch_2": State.NONE,
                    }
                )

    @pytest.mark.parametrize("target_upper", [target_upper for (_, target_upper) in targets])
    @time_machine.travel("2020-07-07 10:54:05")
    def test_branch_datetime_operator_upper_comparison_within_range(self, target_upper):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_upper = target_upper
        self.branch_op.target_lower = None

        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.NONE,
                    "branch_2": State.SKIPPED,
                }
            )

    @pytest.mark.parametrize("target_lower", [target_lower for (target_lower, _) in targets])
    @time_machine.travel("2020-07-07 10:54:05")
    def test_branch_datetime_operator_lower_comparison_within_range(self, target_lower):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_lower = target_lower
        self.branch_op.target_upper = None

        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.NONE,
                    "branch_2": State.SKIPPED,
                }
            )

    @pytest.mark.parametrize("target_upper", [target_upper for (_, target_upper) in targets])
    @time_machine.travel("2020-07-07 12:00:00")
    def test_branch_datetime_operator_upper_comparison_outside_range(self, target_upper):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_upper = target_upper
        self.branch_op.target_lower = None

        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_1", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.SKIPPED,
                    "branch_2": State.NONE,
                }
            )

    @pytest.mark.parametrize("target_lower", [target_lower for (target_lower, _) in targets])
    @time_machine.travel("2020-07-07 09:00:00")
    def test_branch_datetime_operator_lower_comparison_outside_range(self, target_lower):
        """Check BranchDateTimeOperator branch operation"""
        self.branch_op.target_lower = target_lower
        self.branch_op.target_upper = None

        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_1", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.SKIPPED,
                    "branch_2": State.NONE,
                }
            )

    @pytest.mark.parametrize(
        ("target_lower", "target_upper"),
        targets,
    )
    @time_machine.travel("2020-12-01 09:00:00")
    def test_branch_datetime_operator_use_task_logical_date(self, dag_maker, target_lower, target_upper):
        """Check if BranchDateTimeOperator uses task logical date"""
        in_between_date = timezone.datetime(2020, 7, 7, 10, 30, 0)
        self.branch_op.use_task_logical_date = True
        self.dr = dag_maker.create_dagrun(
            run_id="manual_exec_date__",
            start_date=in_between_date,
            logical_date=in_between_date,
            state=State.RUNNING,
            data_interval=DataInterval(in_between_date, in_between_date),
        )

        self.branch_op.target_lower = target_lower
        self.branch_op.target_upper = target_upper
        if AIRFLOW_V_3_0_1:
            from airflow.providers.common.compat.sdk import DownstreamTasksSkipped

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                self.dag_maker.run_ti("datetime_branch", self.dr)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            self.dag_maker.run_ti("datetime_branch", self.dr)

            self._assert_task_ids_match_states(
                {
                    "datetime_branch": State.SUCCESS,
                    "branch_1": State.NONE,
                    "branch_2": State.SKIPPED,
                }
            )

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Skip on Airflow < 3.0")
    @time_machine.travel("2020-12-01 09:00:00")
    def test_choose_branch_should_use_run_after_when_logical_date_none(self, dag_maker):
        with dag_maker(
            "branch_datetime_operator_uses_run_after",
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")

            branch_op = BranchDateTimeOperator(
                task_id="datetime_branch",
                follow_task_ids_if_true="branch_1",
                follow_task_ids_if_false="branch_2",
                target_upper=datetime.datetime(2020, 9, 7, 11, 0, 0),
                target_lower=datetime.datetime(2020, 6, 7, 10, 0, 0),
                use_task_logical_date=True,
            )
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)

        dr = dag_maker.create_dagrun(
            run_id="manual__run_after",
            start_date=DEFAULT_DATE,
            state=State.RUNNING,
            **{"run_after": timezone.datetime(2020, 8, 7)},
        )
        assert branch_op.choose_branch(context={"dag_run": dr}) == "branch_1"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import warnings
from datetime import datetime, timedelta
from time import sleep
from typing import TYPE_CHECKING, Any

from deprecated.classic import deprecated
from packaging.version import Version

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.common.compat.sdk import AirflowSkipException, BaseSensorOperator, conf, timezone
from airflow.providers.standard.triggers.temporal import DateTimeTrigger, TimeDeltaTrigger
from airflow.providers.standard.version_compat import AIRFLOW_V_3_0_PLUS

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


def _get_airflow_version():
    from airflow import __version__ as airflow_version

    return Version(Version(airflow_version).base_version)


class TimeDeltaSensor(BaseSensorOperator):
    """
    Waits for a timedelta.

    The delta will be evaluated against data_interval_end if present for the dag run,
    otherwise run_after will be used.

    :param delta: time to wait before succeeding.
    :param deferrable: Run sensor in deferrable mode. If set to True, task will defer itself to avoid taking up a worker slot while it is waiting.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/operator:TimeDeltaSensor`

    """

    def __init__(
        self,
        *,
        delta: timedelta,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        end_from_trigger: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.delta = delta
        self.deferrable = deferrable
        self.end_from_trigger = end_from_trigger

    def _derive_base_time(self, context: Context) -> datetime:
        """
        Get the "base time" against which the delta should be calculated.

        If data_interval_end is populated, use it; else use run_after.
        """
        data_interval_end = context.get("data_interval_end")
        if data_interval_end:
            if not isinstance(data_interval_end, datetime):
                raise ValueError("`data_interval_end` returned non-datetime object")

            return data_interval_end

        if not data_interval_end and not AIRFLOW_V_3_0_PLUS:
            raise ValueError("`data_interval_end` not found in task context.")

        dag_run = context.get("dag_run")
        if not dag_run:
            raise ValueError("`dag_run` not found in task context")
        return dag_run.run_after

    def poke(self, context: Context) -> bool:
        base_time = self._derive_base_time(context=context)
        target_dttm = base_time + self.delta
        self.log.info("Checking if the delta has elapsed base_time=%s, delta=%s", base_time, self.delta)
        return timezone.utcnow() > target_dttm

    """
    Asynchronous execution
    """

    def execute(self, context: Context) -> Any:
        """
        Depending on the deferrable flag, either execute the sensor in a blocking way or defer it.

        - Sync path → use BaseSensorOperator.execute() which loops over ``poke``.
        - Async path → defer to DateTimeTrigger and free the worker slot.
        """
        if not self.deferrable:
            return super().execute(context=context)

        # Deferrable path
        base_time = self._derive_base_time(context=context)
        target_dttm: datetime = base_time + self.delta

        if timezone.utcnow() > target_dttm:
            # If the target datetime is in the past, return immediately
            return True
        try:
            if AIRFLOW_V_3_0_PLUS:
                trigger = DateTimeTrigger(moment=target_dttm, end_from_trigger=self.end_from_trigger)
            else:
                trigger = DateTimeTrigger(moment=target_dttm)
        except (TypeError, ValueError) as e:
            if self.soft_fail:
                raise AirflowSkipException("Skipping due to soft_fail is set to True.") from e
            raise

        # todo: remove backcompat when min airflow version greater than 2.11
        timeout: int | float | timedelta
        if AIRFLOW_V_3_0_PLUS:
            timeout = self.timeout
        else:
            # <=2.11 requires timedelta
            timeout = timedelta(seconds=self.timeout)

        self.defer(
            trigger=trigger,
            method_name="execute_complete",
            timeout=timeout,
        )

    def execute_complete(self, context: Context, event: Any = None) -> None:
        """Handle the event when the trigger fires and return immediately."""
        return None


# TODO: Remove in the next major release
@deprecated(
    "Use `TimeDeltaSensor` with `deferrable=True` instead", category=AirflowProviderDeprecationWarning
)
class TimeDeltaSensorAsync(TimeDeltaSensor):
    """
    Deprecated. Use TimeDeltaSensor with deferrable=True instead.

    :sphinx-autoapi-skip:
    """

    def __init__(self, *, end_from_trigger: bool = False, delta, **kwargs) -> None:
        warnings.warn(
            "TimeDeltaSensorAsync is deprecated and will be removed in a future version. Use `TimeDeltaSensor` with `deferrable=True` instead.",
            AirflowProviderDeprecationWarning,
            stacklevel=2,
        )
        super().__init__(delta=delta, deferrable=True, end_from_trigger=end_from_trigger, **kwargs)


class WaitSensor(BaseSensorOperator):
    """
    A sensor that waits a specified period of time before completing.

    This differs from TimeDeltaSensor because the time to wait is measured from the start of the task, not
    the data_interval_end of the DAG run.

    :param time_to_wait: time length to wait after the task starts before succeeding.
    :param deferrable: Run sensor in deferrable mode
    """

    def __init__(
        self,
        time_to_wait: timedelta | int,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.deferrable = deferrable
        if isinstance(time_to_wait, int):
            self.time_to_wait = timedelta(minutes=time_to_wait)
        else:
            self.time_to_wait = time_to_wait

    def execute(self, context: Context) -> None:
        if self.deferrable:
            self.defer(
                trigger=(
                    TimeDeltaTrigger(self.time_to_wait, end_from_trigger=True)
                    if AIRFLOW_V_3_0_PLUS
                    else TimeDeltaTrigger(self.time_to_wait)
                ),
                method_name="execute_complete",
            )
        else:
            sleep(int(self.time_to_wait.total_seconds()))
